import pytest
import typing
from dataclasses import dataclass, fields
import copy

from ldf_adapter.results import Failure, FatalError
from ldf_adapter.approval.models import (
    PendingUser,
    PendingGroup,
    PendingMemberships,
    DeploymentState,
)
from ldf_adapter.approval.db.sqlite import (
    pytype_to_sqltype,
    sql_command_create_table,
    sql_command_insert,
    sql_command_update,
    SQLiteConnector,
)


MOCK_USER = PendingUser(
    unique_id="unique_id",
    sub="sub",
    iss="iss",
    email="email@domain",
    full_name="full_name",
    username="username",
    state=DeploymentState.PENDING,
    cmd="cmd",
    infodict={},
)

MOCK_GROUP = PendingGroup(
    name="name",
    state=DeploymentState.PENDING,
    cmd="cmd",
    infodict={},
)

MOCK_MEMBERSHIPS = PendingMemberships(
    unique_id="unique_id",
    supplementary_groups=["group1", "group2"],
    removal_groups=["group3", "group4"],
    state=DeploymentState.PENDING,
    cmd="cmd",
    infodict={},
)


@dataclass
class MyDataModel:
    id: int
    name: str
    age: int
    height: float
    weight: float
    is_married: bool
    children: list
    address: dict
    deployment_state: DeploymentState


MOCK_MY_DATA_MODEL = MyDataModel(
    id=1,
    name="name",
    age=20,
    height=1.8,
    weight=70.0,
    is_married=True,
    children=[],
    address={"street": "street", "city": "city", "zip": "zip"},
    deployment_state=DeploymentState.PENDING,
)

NoneType = None.__class__


@pytest.mark.parametrize(
    "pytype,sqltype",
    [
        (int, "integer"),
        (float, "real"),
        (str, "text"),
        (bytes, "blob"),
        (dict, "dict"),
        (list, "list"),
        (DeploymentState, "DeploymentState"),
        (typing.Dict[int, str], "dict"),
        (typing.List[int], "list"),
        (typing.ByteString, "blob"),
        (typing.Callable, "text"),
        (typing.Any, "text"),
        (typing.Tuple[int, int], "text"),
        (typing.Type, "text"),
        (typing.NamedTuple, "text"),
        (typing.Set[str], "list"),
        (typing.Union[NoneType, int], "integer"),
        (typing.Union[int, NoneType], "integer"),
        (typing.Union[NoneType, int, str], "text"),
        (typing.Optional[int], "integer"),
        (typing.Optional[float], "real"),
        (typing.Optional[str], "text"),
        (typing.Optional[bytes], "blob"),
        (typing.Optional[dict], "dict"),
        (typing.Optional[list], "list"),
        (typing.Optional[DeploymentState], "DeploymentState"),
        (typing.Optional[typing.Dict], "dict"),
        (typing.Optional[typing.List], "list"),
        (typing.Optional[typing.ByteString], "blob"),
        (typing.Optional[typing.Optional[int]], "integer"),
        (type(None), "null"),
        (type("something"), "text"),
        (type(1), "integer"),
        (type(1.0), "real"),
        (type(b"something"), "blob"),
        (type({}), "dict"),
        (type([]), "list"),
        (type(DeploymentState.PENDING), "DeploymentState"),
    ],
)
def test_pytype_to_sqltype(pytype: typing.Type, sqltype: str):
    assert pytype_to_sqltype(pytype) == sqltype


create_my_data_model_table_cmd = """
    create table if not exists my_data_model
    (
        id integer,
        name text,
        age integer,
        height real,
        weight real,
        is_married boolean,
        children list,
        address dict,
        deployment_state DeploymentState,
        primary key (id, address)
    )
"""

create_pending_users_table_cmd = """
    create table if not exists pending_users
    (
        unique_id text,
        sub text,
        iss text,
        email text,
        full_name text,
        username text,
        state DeploymentState,
        cmd text,
        infodict dict,
        primary key (unique_id)
    )
"""

create_pending_groups_table_cmd = """
    create table if not exists pending_groups
    (
        name text,
        state DeploymentState,
        cmd text,
        infodict dict,
        primary key (name)
    )
"""

create_pending_memberships_table_cmd = """
    create table if not exists pending_memberships
    (
        unique_id text,
        supplementary_groups list,
        removal_groups list,
        state DeploymentState,
        cmd text,
        infodict dict,
        primary key (unique_id)
    )
"""


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,expected_cmd",
    [
        (
            MyDataModel,
            "my_data_model",
            ["id", "address"],
            create_my_data_model_table_cmd,
        ),
        (PendingUser, "pending_users", "unique_id", create_pending_users_table_cmd),
        (PendingGroup, "pending_groups", "name", create_pending_groups_table_cmd),
        (
            PendingMemberships,
            "pending_memberships",
            "unique_id",
            create_pending_memberships_table_cmd,
        ),
    ],
)
def test_sql_command_create_table(data_model, table_name, primary_key, expected_cmd):
    assert (
        "".join(
            sql_command_create_table(data_model, table_name, primary_key).split()
        ).lower()
        == "".join(expected_cmd.split()).lower()
    )


insert_my_data_model_cmd = """
    insert into my_data_model
    (id, name, age, height, weight, is_married, children, address, deployment_state)
    values (?,?,?,?,?,?,?,?,?)
"""

insert_pending_users_cmd = """
    insert into pending_users
    (unique_id, sub, iss, email, full_name, username, state, cmd, infodict)
    values (?,?,?,?,?,?,?,?,?)
"""

insert_pending_groups_cmd = """
    insert into pending_groups
    (name, state, cmd, infodict)
    values (?,?,?,?)
"""

insert_pending_memberships_cmd = """
    insert into pending_memberships
    (unique_id, supplementary_groups, removal_groups, state, cmd, infodict)
    values (?,?,?,?,?,?)
"""


@pytest.mark.parametrize(
    "data_model,table_name,expected_cmd",
    [
        (MyDataModel, "my_data_model", insert_my_data_model_cmd),
        (PendingUser, "pending_users", insert_pending_users_cmd),
        (PendingGroup, "pending_groups", insert_pending_groups_cmd),
        (PendingMemberships, "pending_memberships", insert_pending_memberships_cmd),
    ],
)
def test_sql_command_insert(data_model, table_name, expected_cmd):
    assert (
        "".join(sql_command_insert(data_model, table_name).split()).lower()
        == "".join(expected_cmd.split()).lower()
    )


update_my_data_model_cmd = """
    update my_data_model
    set name = ?, age = ?, height = ?, weight = ?
    where id = ? and address = ?
"""

update_pending_users_cmd = """
    update pending_users
    set state = ?
    where unique_id = ?
"""

update_pending_groups_cmd = """
    update pending_groups
    set state = ?, cmd = ?, infodict = ?
    where name = ?
"""

update_pending_memberships_cmd = """
    update pending_memberships
    set unique_id = ?
    where unique_id = ?
"""

update_pending_users_cmd2 = """
    update pending_users
    set unknown_field = ?
    where unique_id = ?
"""

update_pending_users_cmd3 = """
    update pending_users
    set unique_id = ?
    where unknown_key = ?
"""


@pytest.mark.parametrize(
    "table_name,columns,key,expected_cmd",
    [
        (
            "my_data_model",
            ["name", "age", "height", "weight"],
            ["id", "address"],
            update_my_data_model_cmd,
        ),
        ("pending_users", ["state"], "unique_id", update_pending_users_cmd),
        (
            "pending_groups",
            ["state", "cmd", "infodict"],
            ["name"],
            update_pending_groups_cmd,
        ),
        (
            "pending_memberships",
            ["unique_id"],
            "unique_id",
            update_pending_memberships_cmd,
        ),
        ("pending_users", ["unknown_field"], "unique_id", update_pending_users_cmd2),
        ("pending_users", ["unique_id"], "unknown_key", update_pending_users_cmd3),
    ],
)
def test_sql_command_update(table_name, columns, key, expected_cmd):
    assert (
        "".join(sql_command_update(table_name, columns, key).split()).lower()
        == "".join(expected_cmd.split()).lower()
    )


def new_init(self: SQLiteConnector, location: str) -> None:
    self.location = ":memory:"


# data_model, table_name, primary_key, mock_object
sample_data_sqlite = [
    (PendingUser, "pending_users", ["unique_id"], MOCK_USER),
    (PendingGroup, "pending_groups", ["name"], MOCK_GROUP),
    (PendingMemberships, "pending_memberships", ["unique_id"], MOCK_MEMBERSHIPS),
    (MyDataModel, "my_data_model", ["id", "address"], MOCK_MY_DATA_MODEL),
]


@pytest.fixture(scope="function")
def mock_connector():
    SQLiteConnector.__init__ = new_init
    connector = SQLiteConnector("")
    connector.connect()
    yield connector


@pytest.fixture(scope="function")
def mock_connector_with_table(data_model, table_name, primary_key):
    SQLiteConnector.__init__ = new_init
    connector = SQLiteConnector("")
    connector.connect()
    connector.create(data_model, table_name, primary_key)
    yield connector


@pytest.fixture(scope="function")
def mock_connector_with_entry(data_model, table_name, primary_key, mock_object):
    SQLiteConnector.__init__ = new_init
    connector = SQLiteConnector("")
    connector.connect()
    connector.create(data_model, table_name, primary_key)
    entry = tuple(getattr(mock_object, field.name) for field in fields(data_model))
    connector.insert(data_model, table_name, entry)
    yield connector


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object", sample_data_sqlite
)
def test_create_unknown_key(
    mock_connector, data_model, table_name, primary_key, mock_object
):
    with pytest.raises(SystemExit) as excinfo:
        mock_connector.create(data_model, table_name, "unknown_key")
    assert excinfo.value.code == FatalError.exit_code


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object", sample_data_sqlite
)
def test_create_successful(
    mock_connector, data_model, table_name, primary_key, mock_object
):
    mock_connector.create(data_model, table_name, primary_key)
    c = mock_connector.connection.cursor()
    # check if table exists
    c.execute(
        f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}';"
    )
    assert c.fetchone()[0] == table_name
    # check if table has the correct columns
    c.execute(f"PRAGMA table_info({table_name});")
    columns = [row[1] for row in c.fetchall()]
    assert columns == [f.name for f in fields(data_model)]
    # check if table has the correct primary key
    c.execute(f"PRAGMA table_info({table_name});")
    primary_key_set = [row[1] for row in c.fetchall() if row[5] != 0]
    assert primary_key_set == primary_key


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object", sample_data_sqlite
)
def test_create_if_not_exists(
    mock_connector, data_model, table_name, primary_key, mock_object
):
    mock_connector.create(data_model, table_name, primary_key)
    mock_connector.create(data_model, table_name, primary_key)
    # check that second create didn't raise any exception, but didn't create a new table
    result = (
        mock_connector.connection.cursor()
        .execute(
            f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}';"
        )
        .fetchall()
    )
    assert len(result) == 1
    assert result[0][0] == table_name


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object", sample_data_sqlite
)
def test_insert_table_doesnt_exist(
    mock_connector, data_model, table_name, primary_key, mock_object
):
    entry = tuple(getattr(mock_object, field.name) for field in fields(data_model))
    with pytest.raises(Failure):
        mock_connector.insert(data_model, table_name, entry)


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object", sample_data_sqlite
)
def test_insert_wrong_number_of_columns(
    mock_connector_with_table, data_model, table_name, primary_key, mock_object
):
    entry = tuple(getattr(mock_object, field.name) for field in fields(data_model))
    wrong_entry = entry + ("wrong",)
    with pytest.raises(Failure):
        mock_connector_with_table.insert(data_model, table_name, wrong_entry)


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object", sample_data_sqlite
)
def test_insert_successful(
    mock_connector_with_table, data_model, table_name, primary_key, mock_object
):
    entry = tuple(getattr(mock_object, field.name) for field in fields(data_model))
    assert mock_connector_with_table.insert(data_model, table_name, entry) == True
    # check that inserted entry is the same as the mock object
    result = (
        mock_connector_with_table.connection.cursor()
        .execute(f"SELECT * FROM {table_name};")
        .fetchall()
    )
    assert len(result) == 1
    assert result[0] == entry


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object", sample_data_sqlite
)
def test_insert_entry_exists(
    mock_connector_with_entry, data_model, table_name, primary_key, mock_object
):
    entry = tuple(getattr(mock_object, field.name) for field in fields(data_model))
    assert mock_connector_with_entry.insert(data_model, table_name, entry) == False


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object", sample_data_sqlite
)
def test_select_table_doesnt_exist(
    mock_connector, data_model, table_name, primary_key, mock_object
):
    if isinstance(primary_key, list):
        primary_key_value = tuple([getattr(mock_object, key) for key in primary_key])
    else:
        primary_key_value = (getattr(mock_object, primary_key),)
    with pytest.raises(Failure):
        mock_connector.select(data_model, table_name, primary_key, primary_key_value)


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object", sample_data_sqlite
)
def test_select_entry_not_found(
    mock_connector_with_table, data_model, table_name, primary_key, mock_object
):
    if isinstance(primary_key, list):
        primary_key_value = tuple([getattr(mock_object, key) for key in primary_key])
    else:
        primary_key_value = (getattr(mock_object, primary_key),)
    assert (
        mock_connector_with_table.select(
            data_model, table_name, primary_key, primary_key_value
        )
        is None
    )


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object", sample_data_sqlite
)
def test_select_found(
    mock_connector_with_entry, data_model, table_name, primary_key, mock_object
):
    if isinstance(primary_key, list):
        primary_key_value = tuple([getattr(mock_object, key) for key in primary_key])
    else:
        primary_key_value = (getattr(mock_object, primary_key),)
    assert (
        mock_connector_with_entry.select(
            data_model, table_name, primary_key, primary_key_value
        )
        == mock_object
    )


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object", sample_data_sqlite
)
def test_delete_table_doesnt_exist(
    mock_connector, data_model, table_name, primary_key, mock_object
):
    if isinstance(primary_key, list):
        primary_key_value = tuple([getattr(mock_object, key) for key in primary_key])
    else:
        primary_key_value = (getattr(mock_object, primary_key),)
    with pytest.raises(Failure):
        mock_connector.delete(table_name, primary_key, primary_key_value)


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object", sample_data_sqlite
)
def test_delete_entry_doesnt_exist(
    mock_connector_with_table, data_model, table_name, primary_key, mock_object
):
    if isinstance(primary_key, list):
        primary_key_value = tuple([getattr(mock_object, key) for key in primary_key])
    else:
        primary_key_value = (getattr(mock_object, primary_key),)
    mock_connector_with_table.delete(table_name, primary_key, primary_key_value)


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object", sample_data_sqlite
)
def test_delete_successful(
    mock_connector_with_entry, data_model, table_name, primary_key, mock_object
):
    if isinstance(primary_key, list):
        primary_key_value = tuple([getattr(mock_object, key) for key in primary_key])
    else:
        primary_key_value = (getattr(mock_object, primary_key),)
    mock_connector_with_entry.delete(table_name, primary_key, primary_key_value)
    result = (
        mock_connector_with_entry.connection.cursor()
        .execute(f"SELECT * FROM {table_name};")
        .fetchall()
    )
    assert len(result) == 0


# "data_model,table_name,primary_key,mock_object,column,value,second_value"
sample_data_sqlite_select_on_column = [
    (
        PendingUser,
        "pending_users",
        ["unique_id"],
        MOCK_USER,
        "cmd",
        "cmd",
        "another_unique_id",
    ),
    (
        PendingGroup,
        "pending_groups",
        ["name"],
        MOCK_GROUP,
        "cmd",
        "cmd",
        "another_name",
    ),
    (
        PendingMemberships,
        "pending_memberships",
        ["unique_id"],
        MOCK_MEMBERSHIPS,
        "cmd",
        "cmd",
        "another_unique_id",
    ),
    (
        MyDataModel,
        "my_data_model",
        ["id", "address"],
        MOCK_MY_DATA_MODEL,
        "name",
        "name",
        "another_id",
    ),
]


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object,column,value,second_value",
    sample_data_sqlite_select_on_column,
)
def test_select_on_column_table_doesnt_exist(
    mock_connector,
    data_model,
    table_name,
    primary_key,
    mock_object,
    column,
    value,
    second_value,
):
    with pytest.raises(Failure):
        mock_connector.select(data_model, table_name, column, (value,))


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object,column,value,second_value",
    sample_data_sqlite_select_on_column,
)
def test_select_on_column_not_found(
    mock_connector_with_table,
    data_model,
    table_name,
    primary_key,
    mock_object,
    column,
    value,
    second_value,
):
    assert (
        mock_connector_with_table.select(data_model, table_name, column, (value,))
        is None
    )


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object,column,value,second_value",
    sample_data_sqlite_select_on_column,
)
def test_select_on_column_found(
    mock_connector_with_entry,
    data_model,
    table_name,
    primary_key,
    mock_object,
    column,
    value,
    second_value,
):
    assert (
        mock_connector_with_entry.select(data_model, table_name, column, (value,))
        == mock_object
    )


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object,column,value,second_value",
    sample_data_sqlite_select_on_column,
)
def test_select_on_column_multiple_entries(
    mock_connector_with_entry,
    data_model,
    table_name,
    primary_key,
    mock_object,
    column,
    value,
    second_value,
):
    # insert a second entry with same column value
    mock_object2 = copy.copy(mock_object)
    mock_object2.__setattr__(
        primary_key[0] if isinstance(primary_key, list) else primary_key, second_value
    )
    second_entry = tuple(
        getattr(mock_object2, field.name) for field in fields(data_model)
    )
    mock_connector_with_entry.insert(data_model, table_name, second_entry)
    # check that select returns first entry when multiple entries are found
    assert (
        mock_connector_with_entry.select(data_model, table_name, column, (value,))
        == mock_object
    )


# "data_model,table_name,primary_key,mock_object,column,value"
sample_data_sqlite_update = [
    (PendingUser, "pending_users", "unique_id", MOCK_USER, "cmd", "cmd2"),
    (PendingGroup, "pending_groups", "name", MOCK_GROUP, "cmd", "cmd2"),
    (
        PendingMemberships,
        "pending_memberships",
        "unique_id",
        MOCK_MEMBERSHIPS,
        "cmd",
        "cmd2",
    ),
    (
        MyDataModel,
        "my_data_model",
        ["id", "address"],
        MOCK_MY_DATA_MODEL,
        "name",
        "name2",
    ),
]


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object,column,value",
    sample_data_sqlite_update,
)
def test_update_table_doesnt_exist(
    mock_connector, data_model, table_name, primary_key, mock_object, column, value
):
    if isinstance(primary_key, list):
        primary_key_value = tuple([getattr(mock_object, key) for key in primary_key])
    else:
        primary_key_value = (getattr(mock_object, primary_key),)
    entry = (value,) + primary_key_value
    with pytest.raises(Failure):
        mock_connector.update(table_name, [column], primary_key, entry)


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object,column,value",
    sample_data_sqlite_update,
)
def test_update_entry_doesnt_exist(
    mock_connector_with_table,
    data_model,
    table_name,
    primary_key,
    mock_object,
    column,
    value,
):
    if isinstance(primary_key, list):
        primary_key_value = tuple([getattr(mock_object, key) for key in primary_key])
    else:
        primary_key_value = (getattr(mock_object, primary_key),)
    entry = (value,) + primary_key_value
    # does not fail, but does not modify anything
    result_before = (
        mock_connector_with_table.connection.cursor()
        .execute(f"SELECT * FROM {table_name};")
        .fetchall()
    )
    mock_connector_with_table.update(table_name, [column], primary_key, entry)
    # check that nothing changed
    result_after = (
        mock_connector_with_table.connection.cursor()
        .execute(f"SELECT * FROM {table_name};")
        .fetchall()
    )
    assert result_before == result_after


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object,column,value",
    sample_data_sqlite_update,
)
def test_update_wrong_number_of_columns(
    mock_connector_with_entry,
    data_model,
    table_name,
    primary_key,
    mock_object,
    column,
    value,
):
    if isinstance(primary_key, list):
        primary_key_value = tuple([getattr(mock_object, key) for key in primary_key])
    else:
        primary_key_value = (getattr(mock_object, primary_key),)
    entry = (value,) + primary_key_value
    with pytest.raises(Failure):
        mock_connector_with_entry.update(
            table_name, [column, column], primary_key, entry
        )


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object,column,value",
    sample_data_sqlite_update,
)
def test_update_wrong_number_of_values(
    mock_connector_with_entry,
    data_model,
    table_name,
    primary_key,
    mock_object,
    column,
    value,
):
    if isinstance(primary_key, list):
        primary_key_value = tuple([getattr(mock_object, key) for key in primary_key])
    else:
        primary_key_value = (getattr(mock_object, primary_key),)
    entry = (value,) + primary_key_value
    with pytest.raises(Failure):
        mock_connector_with_entry.update(table_name, [column], primary_key, entry[:-1])


@pytest.mark.parametrize(
    "data_model,table_name,primary_key,mock_object,column,value",
    sample_data_sqlite_update,
)
def test_update_successful(
    mock_connector_with_entry,
    data_model,
    table_name,
    primary_key,
    mock_object,
    column,
    value,
):
    if isinstance(primary_key, list):
        primary_key_value = tuple([getattr(mock_object, key) for key in primary_key])
    else:
        primary_key_value = (getattr(mock_object, primary_key),)
    entry = (value,) + primary_key_value
    mock_connector_with_entry.update(table_name, [column], primary_key, entry)
    result = (
        mock_connector_with_entry.connection.cursor()
        .execute(f"SELECT {column} FROM {table_name};")
        .fetchall()
    )
    assert len(result) == 1
    assert result[0][0] == value



"""Parse commandline options"""

#
# Author: Marcus Hardt <hardt@kit.edu>
#
# pylint # {{{
# vim: tw=100 foldmethod=indent
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=redefined-outer-name
# pylint: disable=missing-docstring, trailing-whitespace, trailing-newlines, too-few-public-methods
# }}}
import os
import sys
import logging
import argparse

logger = logging.getLogger(__name__)


def parseOptions():
    """Parse the commandline options"""

    logger.info("reading config")
    path_of_executable = os.path.realpath(sys.argv[0])
    folder_of_executable = os.path.split(path_of_executable)[0]
    full_name_of_executable = os.path.split(path_of_executable)[1]
    name_of_executable = full_name_of_executable.rstrip(".py")
    try:
        config_in_home = os.environ["HOME"] + "/.config/%s.conf" % name_of_executable
    except KeyError:
        pass

    parser = argparse.ArgumentParser(description=name_of_executable)

    parser.add_argument(
        "--config_file",
        "--config",
        "-c",
        "--conf",
        default="/etc/feudal/feudal_adapter.conf",
        help="Default: /etc/feudal/feudal_adapter.conf",
    )

    parser.add_argument(
        "--test", action="store_true", help="Test notifier configuration"
    )

    return parser


# reparse args on import
args = parseOptions().parse_args()



import pytest
from attr import dataclass
import pytest
import json
from typing import List
import logging

import ldf_adapter
from . import settings

logger = logging.getLogger(__name__)


@dataclass
class DbUser:
    unique_id: str
    username: str
    primary_group: str
    groups: List[str]
    pending: bool = False
    rejected: bool = False
    suspended: bool = False
    limited: bool = False


@dataclass
class DbGroup:
    name: str
    members: List[str]


class MockDB:
    """Simple user db represented as a dict"""

    def __init__(self):
        self.users = {}
        self.groups = {}

    def reset(self):
        self.users = {}
        self.groups = {}

    def add_user(self, **kwargs):
        self.users[kwargs["unique_id"]] = DbUser(**kwargs)

    def add_group(self, **kwargs):
        self.groups[kwargs["name"]] = DbGroup(**kwargs)

    def has_user(self, unique_id):
        return unique_id in self.users

    def name_taken(self, username):
        return any(user.username == username for user in self.users.values())

    def has_group(self, name):
        return name in self.groups

    def delete_user(self, unique_id):
        del self.users[unique_id]


MOCK_DB = MockDB()


class MockBackendUser:
    """Mock user for the backend"""

    def __init__(self, userinfo):
        self.unique_id = userinfo.unique_id
        if self.exists():
            self.set_username(self.get_username())
            self.primary_group = MockBackendGroup(self.get_primary_group())
        else:
            self.set_username(userinfo.username)
            self.primary_group = MockBackendGroup(userinfo.primary_group)

    def exists(self):
        return MOCK_DB.has_user(unique_id=self.unique_id)

    def name_taken(self, name):
        return MOCK_DB.name_taken(name)

    def create(self):
        MOCK_DB.add_user(
            unique_id=self.unique_id,
            username=self.username,
            primary_group=self.primary_group.name,
            groups=[self.primary_group.name],
        )
        MOCK_DB.groups[self.primary_group.name].members.append(self.username)

    def create_tostring(self):
        return json.dumps(
            {
                "unique_id": self.unique_id,
                "username": self.username,
                "primary_group": self.primary_group.name,
                "groups": [self.primary_group.name],
            }
        )

    @staticmethod
    def create_fromstring(create_cmd):
        kwargs = json.loads(create_cmd)
        MOCK_DB.add_user(**kwargs)
        MOCK_DB.groups[kwargs["primary_group"]].members.append(kwargs["username"])

    def update(self):
        pass

    def delete(self):
        MOCK_DB.delete_user(self.unique_id)

    def mod(self, supplementary_groups=None):
        if supplementary_groups is None:
            supplementary_groups_names = []
        else:
            supplementary_groups_names = [group.name for group in supplementary_groups]
        current_groups = self.get_groups()
        groups_to_add = list(set(supplementary_groups_names) - set(current_groups))
        groups_to_remove = list(set(current_groups) - set(supplementary_groups_names))

        MOCK_DB.users[self.unique_id].groups = supplementary_groups_names
        for group in groups_to_add:
            MOCK_DB.groups[group].members.append(self.username)
        for group in groups_to_remove:
            MOCK_DB.groups[group].members.remove(self.username)
        return groups_to_add, groups_to_remove

    def get_groups(self):
        return MOCK_DB.users[self.unique_id].groups

    def install_ssh_keys(self):
        pass

    def uninstall_ssh_keys(self):
        pass

    def get_username(self):
        try:
            return MOCK_DB.users[self.unique_id].username
        except KeyError:
            return None

    def set_username(self, username):
        self.username = username

    def get_primary_group(self):
        return MOCK_DB.users[self.unique_id].primary_group

    @property
    def credentials(self):
        return {}

    def is_rejected(self):
        return MOCK_DB.users[self.unique_id].rejected

    def is_suspended(self):
        return MOCK_DB.users[self.unique_id].suspended

    def is_pending(self):
        return MOCK_DB.users[self.unique_id].pending

    def is_limited(self):
        return MOCK_DB.users[self.unique_id].limited

    def suspend(self):
        MOCK_DB.users[self.unique_id].suspended = True

    def resume(self):
        MOCK_DB.users[self.unique_id].suspended = False

    def limit(self):
        MOCK_DB.users[self.unique_id].limited = True

    def unlimit(self):
        MOCK_DB.users[self.unique_id].limited = False

    def execute(self, hook, *hook_args):
        pass


class MockBackendGroup:
    def __init__(self, name):
        self.name = name

    def exists(self):
        return MOCK_DB.has_group(name=self.name)

    @property
    def members(self):
        return MOCK_DB.groups[self.name].members

    def create(self):
        MOCK_DB.add_group(name=self.name, members=[])

    def create_tostring(self):
        return json.dumps({"name": self.name, "members": []})

    @staticmethod
    def create_fromstring(create_cmd):
        kwargs = json.loads(create_cmd)
        MOCK_DB.add_group(**kwargs)


@pytest.fixture(scope="function")
def user(data, monkeypatch):
    """Creates a User from provided dict"""
    with monkeypatch.context() as mp:
        mp.setattr("ldf_adapter.backend.User", MockBackendUser)
        mp.setattr("ldf_adapter.backend.Group", MockBackendGroup)
        MOCK_DB.reset()
        user = ldf_adapter.User(data)
        yield user


@pytest.mark.parametrize(
    "data,assurance_verified",
    [
        (settings.INPUT_EGI, True),
        (settings.INPUT_UNITY, True),
        (settings.INPUT_DEEP_IAM, False),
        (settings.INPUT_INDIGO_IAM, False),
        (settings.INPUT_KIT, False),
    ],
)
def test_assurance_verifier(user, assurance_verified):
    assert user.assurance_verifier()(user.data.assurance) == assurance_verified


@pytest.mark.parametrize("data", settings.ALL_INPUT)
def test_get_status(user):
    assert user.get_status().attributes == {
        "state": "not_deployed",
        "message": "No message",
    }


@pytest.mark.skip(
    """Can't get monkeypatch to remove the default group,
hence the SystemExit exception is never raised"""
)
@pytest.mark.parametrize("data", [settings.INPUT_EGI])
def test_deploy_fails_no_primary_group(user, monkeypatch):
    """Make sure we raise an error, if no group is passed _AND_ there's no
    fallback_group configured"""
    #  attr = monkeypatch.__getattribute__("ldf_adapter.backend.local_unix.CONFIG.ldf_adapter.fallback_group")
    #  logger.error(F"monkeypatch: {attr}")
    #  for a in attr:
    #      logger.error(F"monkeypatch: {a}")
    monkeypatch.setattr("ldf_adapter.CONFIG.messages.log_level", "DEBUG")
    monkeypatch.setattr("ldf_adapter.userinfo.CONFIG.messages.log_level", "DEBUG")
    monkeypatch.setattr(
        "ldf_adapter.backend.local_unix.CONFIG.messages.log_level", "DEBUG"
    )
    monkeypatch.delattr(
        "ldf_adapter.backend.local_unix.CONFIG.ldf_adapter.fallback_group"
    )
    monkeypatch.delattr("ldf_adapter.userinfo.primary_group")
    #  monkeypatch.delattr("ldf_adapter.userinfo.user.primary_group")
    #  monkeypatch.delattr("ldf_adapter.userinfo.CONFIG.ldf_adapter.fallback_group")
    with pytest.raises(SystemExit):
        user.deploy()


@pytest.mark.parametrize(
    "data,username",
    [
        (settings.INPUT_UNITY, "marhar"),
        (settings.INPUT_DEEP_IAM, "marhar"),
        (settings.INPUT_INDIGO_IAM, "marhar"),
        (settings.INPUT_KIT, "lo0018"),
    ],
)
def test_deploy_name_taken(user, username, monkeypatch):
    monkeypatch.setattr(
        MockBackendUser, "name_taken", lambda x, n: True if n == "marcus" else False
    )
    assert user.deploy().attributes["state"] == "deployed"
    assert user.get_status().attributes == {
        "state": "deployed",
        "message": f"username {username}",
    }


@pytest.mark.parametrize(
    "data,username",
    [
        (settings.INPUT_UNITY, "marcus"),
        (settings.INPUT_DEEP_IAM, "marcus"),
        (settings.INPUT_INDIGO_IAM, "marcus"),
        (settings.INPUT_KIT, "lo0018"),
    ],
)
def test_deploy(user, username):
    result = user.deploy()
    assert result.attributes["state"] == "deployed"
    assert result.attributes["message"].startswith("User was created")
    assert user.get_status().attributes == {
        "state": "deployed",
        "message": f"username {username}",
    }


# test deploy_pooled_user
# test deploy groups
# test deploy group changes


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_suspend(user):
    user.deploy()
    assert user.suspend().attributes["state"] == "suspended"
    assert user.get_status().attributes["state"] == "suspended"


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_suspend_not_deployed(user):
    assert user.suspend().attributes["state"] == "not_deployed"
    assert user.get_status().attributes["state"] == "not_deployed"


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_resume_after_suspend(user):
    user.deploy()
    user.suspend()
    assert user.resume().attributes["state"] == "deployed"
    assert user.get_status().attributes["state"] == "deployed"


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_resume_not_suspended(user):
    user.deploy()
    assert user.resume().attributes["state"] == "deployed"
    assert user.get_status().attributes["state"] == "deployed"


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_suspend_after_limit(user):
    user.deploy()
    user.limit()
    assert user.suspend().attributes["state"] == "suspended"
    assert user.get_status().attributes["state"] == "suspended"


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_resume_after_limit(user):
    user.deploy()
    user.limit()
    assert user.resume().attributes["state"] == "limited"
    assert user.get_status().attributes["state"] == "limited"


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_limit(user):
    user.deploy()
    assert user.limit().attributes["state"] == "limited"
    assert user.get_status().attributes["state"] == "limited"


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_limit_not_deployed(user):
    assert user.limit().attributes["state"] == "not_deployed"
    assert user.get_status().attributes["state"] == "not_deployed"


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_unlimit(user):
    user.deploy()
    user.limit()
    assert user.unlimit().attributes["state"] == "deployed"
    assert user.get_status().attributes["state"] == "deployed"


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_unlimit_not_limited(user):
    user.deploy()
    assert user.unlimit().attributes["state"] == "deployed"
    assert user.get_status().attributes["state"] == "deployed"


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_limit_after_suspend(user):
    user.deploy()
    user.suspend()
    assert user.limit().attributes["state"] == "suspended"
    assert user.get_status().attributes["state"] == "suspended"


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_unlimit_after_limit_and_suspend(user):
    user.deploy()
    user.suspend()
    user.limit()
    assert user.unlimit().attributes["state"] == "suspended"
    assert user.get_status().attributes["state"] == "suspended"


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_undeploy_doesnt_exist(user):
    result = user.undeploy()
    assert result.attributes["state"] == "not_deployed"
    assert result.attributes["message"].startswith("No user for")
    assert user.get_status().attributes["state"] == "not_deployed"


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_undeploy(user):
    user.deploy()
    result = user.undeploy()
    assert result.attributes["state"] == "not_deployed"
    assert result.attributes["message"].endswith("was removed.")
    assert user.get_status().attributes["state"] == "not_deployed"


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_undeploy_after_suspend(user):
    user.deploy()
    user.suspend()
    assert user.undeploy().attributes["state"] == "not_deployed"
    assert user.get_status().attributes["state"] == "not_deployed"


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_undeploy_after_limit(user):
    user.deploy()
    user.limit()
    assert user.undeploy().attributes["state"] == "not_deployed"
    assert user.get_status().attributes["state"] == "not_deployed"


@pytest.mark.parametrize(
    "data",
    [
        settings.INPUT_UNITY,
        settings.INPUT_DEEP_IAM,
        settings.INPUT_INDIGO_IAM,
        settings.INPUT_KIT,
    ],
)
def test_deploy_exists(user):
    user.deploy()
    result = user.deploy()
    assert result.attributes["state"] == "deployed"
    assert result.attributes["message"].startswith("User already existed")
    assert user.get_status().attributes["state"] == "deployed"



name = "eduperson"

import logging

import regex

logger = logging.getLogger(__name__)


class Entitlement:
    """EduPerson Entitlement attribute (de-)serialisation.

    As specified in: https://aarc-project.eu/guidelines/aarc-g002/
    """

    # This regex is not compatible with stdlib 're', we need 'regex'!
    # (because of repeated captures, see https://bugs.python.org/issue7132)
    re = regex.compile(
        r"urn:"
        + r"(?P<nid>[^:]+):(?P<delegated_namespace>[^:]+)"  # Namespace-ID and delegated URN namespace
        + r"(:(?P<subnamespace>[^:]+))*?"  # Sub-namespaces
        + r":group:"
        + r"(?P<group>[^:]+)"  # Root group
        + r"(:(?P<subgroup>[^:]+))*?"  # Sub-groups
        + r"(:role=(?P<role>.+))?"  # Role of the user in the deepest group
        + r"#(?P<group_authority>.+)"  # Authoritative soruce of the entitlement (URN)
    )

    def __init__(self, raw):
        """Parse a raw EduPerson entitlement string in the AARC-G002 format."""
        match = self.re.fullmatch(raw)

        if not match:
            raise ValueError("Invalid entitlement attribute")

        [self.namespace_id] = match.captures("nid")
        [self.delegated_namespace] = match.captures("delegated_namespace")

        self.subnamespaces = match.captures("subnamespace")

        [self.group] = match.captures("group")
        self.subgroups = match.captures("subgroup")
        [self.role] = match.captures("role") or [None]

        [self.group_authority] = match.captures("group_authority")

    def __repr__(self):
        """Serialize the entitlement to the AARC-G002 format.

        This is the inverse to `__init__` and thus `ent_str == repr(Entitlement(ent_str))`
        holds for any valid entitlement.
        """
        return (
            "urn:{namespace_id}:{delegated_namespace}{subnamespaces}"
            + ":group:{group}{subgroups}{role}"
            + "#{group_authority}"
        ).format(
            **{
                **self.__dict__,
                **{
                    "subnamespaces": "".join(
                        [":{}".format(ns) for ns in self.subnamespaces]
                    ),
                    "subgroups": "".join([":{}".format(grp) for grp in self.subgroups]),
                    "role": ":role={}".format(self.role) if self.role else "",
                },
            }
        )

    def __str__(self):
        """Return the entitlement in human-readable string form."""
        return (
            "<Entitlement"
            + " namespace={namespace_id}:{delegated_namespace}{subnamespaces}"
            + " group={group}{subgroups}"
            + "{role}"
            + " auth={group_authority}>"
        ).format(
            **{
                **self.__dict__,
                **{
                    "subnamespaces": "".join(
                        [",{}".format(ns) for ns in self.subnamespaces]
                    ),
                    "subgroups": "".join([",{}".format(grp) for grp in self.subgroups]),
                    "role": " role={}".format(self.role) if self.role else "",
                },
            }
        )

    @property
    def all_toplevel_groups(self):
        return ["_".join([self.group])]

    @property
    def all_groups(self):
        return ["_".join([self.group] + self.subgroups)]

    @property
    def full_namespace(self):
        return [self.delegated_namespace] + self.subnamespaces



"""test name_generators"""

import pytest
import mock

from ldf_adapter import name_generators

marcus_userinfo_json = {
    "state_target": "deployed",
    "user": {
        "userinfo": {
            "acr": "https://aai.egi.eu/LoA#Substantial",
            "eduperson_assurance": ["https://aai.egi.eu/LoA#Substantial"],
            "eduperson_entitlement": [
                "urn:mace:egi.eu:aai.egi.eu:admins:owner@saps-vo.i3m.upv.es",
                "urn:mace:egi.eu:group:EOServices-vo.indra.es:admins:role=owner#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:vm_operator@saps-vo.i3m.upv.es",
                "urn:mace:egi.eu:aai.egi.eu:admins:member@mteam.data.kit.edu",
                "urn:mace:egi.eu:aai.egi.eu:vm_operator@cryoem.instruct-eric.eu",
                "urn:mace:egi.eu:group:covid19.eosc-synergy.eu:role=vm_operator#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:vm_operator@worsica.vo.incd.pt",
                "urn:mace:egi.eu:aai.egi.eu:admins:member@mswss.ui.savba.sk",
                "urn:mace:egi.eu:aai.egi.eu:member@umsa.cerit-sc.cz",
                "urn:mace:egi.eu:aai.egi.eu:vm_operator@o3as.data.kit.edu",
                "urn:mace:egi.eu:group:eosc-synergy.eu:admins:role=member#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:member@covid19.eosc-synergy.eu",
                "urn:mace:egi.eu:aai.egi.eu:admins:member@umsa.cerit-sc.cz",
                "urn:mace:egi.eu:group:covid19.eosc-synergy.eu:role=member#aai.egi.eu",
                "urn:mace:egi.eu:group:cryoem.instruct-eric.eu:admins:role=member#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:vm_operator@umsa.cerit-sc.cz",
                "urn:mace:egi.eu:group:umsa.cerit-sc.cz:admins:role=member#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:admins:member@o3as.data.kit.edu",
                "urn:mace:egi.eu:aai.egi.eu:vm_operator@mswss.ui.savba.sk",
                "urn:mace:egi.eu:group:mteam.data.kit.edu:role=member#aai.egi.eu",
                "urn:mace:egi.eu:group:saps-vo.i3m.upv.es:admins:role=owner#aai.egi.eu",
                "urn:mace:egi.eu:group:umsa.cerit-sc.cz:role=vm_operator#aai.egi.eu",
                "urn:mace:egi.eu:group:goc.egi.eu:role=vm_operator#aai.egi.eu",
                "urn:mace:egi.eu:group:goc.egi.eu:role=member#aai.egi.eu",
                "urn:mace:egi.eu:group:saps-vo.i3m.upv.es:role=vm_operator#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:member@worsica.vo.incd.pt",
                "urn:mace:egi.eu:aai.egi.eu:vm_operator@covid19.eosc-synergy.eu",
                "urn:mace:egi.eu:group:cryoem.instruct-eric.eu:role=vm_operator#aai.egi.eu",
                "urn:mace:egi.eu:group:EOServices-vo.indra.es:admins:role=member#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:member@EOServices-vo.indra.es",
                "urn:mace:egi.eu:group:covid19.eosc-synergy.eu:admins:role=member#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:owner@perfmon",
                "urn:mace:egi.eu:group:o3as.data.kit.edu:admins:role=member#aai.egi.eu",
                "urn:mace:egi.eu:group:registry:perfmon:role=owner#aai.egi.eu",
                "urn:mace:egi.eu:group:worsica.vo.incd.pt:role=member#aai.egi.eu",
                "urn:mace:egi.eu:group:eosc-synergy.eu:role=vm_operator#aai.egi.eu",
                "urn:mace:egi.eu:group:cryoem.instruct-eric.eu:admins:role=owner#aai.egi.eu",
                "urn:mace:egi.eu:group:saps-vo.i3m.upv.es:admins:role=member#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:member@goc.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:member@o3as.data.kit.edu",
                "urn:mace:egi.eu:aai.egi.eu:admins:member@cryoem.instruct-eric.eu",
                "urn:mace:egi.eu:group:umsa.cerit-sc.cz:role=member#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:member@perfmon.m.d.k.e",
                "urn:mace:egi.eu:aai.egi.eu:admins:owner@eosc-synergy.eu",
                "urn:mace:egi.eu:aai.egi.eu:member@saps-vo.i3m.upv.es",
                "urn:mace:egi.eu:group:o3as.data.kit.edu:admins:role=owner#aai.egi.eu",
                "urn:mace:egi.eu:group:mswss.ui.savba.sk:role=member#aai.egi.eu",
                "urn:mace:egi.eu:group:saps-vo.i3m.upv.es:role=member#aai.egi.eu",
                "urn:mace:egi.eu:group:registry:perfmon:role=member#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:vm_operator@eosc-synergy.eu",
                "urn:mace:egi.eu:aai.egi.eu:admins:member@covid19.eosc-synergy.eu",
                "urn:mace:egi.eu:group:EOServices-vo.indra.es:role=member#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:vm_operator@mteam.data.kit.edu",
                "urn:mace:egi.eu:group:worsica.vo.incd.pt:role=vm_operator#aai.egi.eu",
                "urn:mace:egi.eu:group:o3as.data.kit.edu:role=member#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:member@mteam.data.kit.edu",
                "urn:mace:egi.eu:aai.egi.eu:admins:owner@covid19.eosc-synergy.eu",
                "urn:mace:egi.eu:group:mteam.data.kit.edu:perfmon.m.d.k.e:admins:role=member#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:admins:member@EOServices-vo.indra.es",
                "urn:mace:egi.eu:group:mswss.ui.savba.sk:role=vm_operator#aai.egi.eu",
                "urn:mace:egi.eu:group:mteam.data.kit.edu:role=vm_operator#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:admins:owner@mswss.ui.savba.sk",
                "urn:mace:egi.eu:group:eosc-synergy.eu:role=member#aai.egi.eu",
                "urn:mace:egi.eu:group:EOServices-vo.indra.es:role=vm_operator#aai.egi.eu",
                "urn:mace:egi.eu:group:mswss.ui.savba.sk:admins:role=member#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:member@eosc-synergy.eu",
                "urn:mace:egi.eu:aai.egi.eu:admins:owner@umsa.cerit-sc.cz",
                "urn:mace:egi.eu:aai.egi.eu:admins:member@saps-vo.i3m.upv.es",
                "urn:mace:egi.eu:aai.egi.eu:vm_operator@EOServices-vo.indra.es",
                "urn:mace:egi.eu:aai.egi.eu:vm_operator@perfmon.m.d.k.e",
                "urn:mace:egi.eu:aai.egi.eu:admins:owner@EOServices-vo.indra.es",
                "urn:mace:egi.eu:aai.egi.eu:admins:member@eosc-synergy.eu",
                "urn:mace:egi.eu:group:umsa.cerit-sc.cz:admins:role=owner#aai.egi.eu",
                "urn:mace:egi.eu:group:o3as.data.kit.edu:role=vm_operator#aai.egi.eu",
                "urn:mace:egi.eu:group:covid19.eosc-synergy.eu:admins:role=owner#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:vm_operator@goc.egi.eu",
                "urn:mace:egi.eu:group:cryoem.instruct-eric.eu:role=member#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:admins:owner@cryoem.instruct-eric.eu",
                "urn:mace:egi.eu:aai.egi.eu:member@perfmon",
                "urn:mace:egi.eu:aai.egi.eu:admins:owner@mteam.data.kit.edu",
                "urn:mace:egi.eu:group:eosc-synergy.eu:admins:role=owner#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:admins:owner@o3as.data.kit.edu",
                "urn:mace:egi.eu:group:mteam.data.kit.edu:perfmon.m.d.k.e:role=member#aai.egi.eu",
                "urn:mace:egi.eu:group:mteam.data.kit.edu:admins:role=owner#aai.egi.eu",
                "urn:mace:egi.eu:group:mswss.ui.savba.sk:admins:role=owner#aai.egi.eu",
                "urn:mace:egi.eu:group:mteam.data.kit.edu:admins:role=member#aai.egi.eu",
                "urn:mace:egi.eu:aai.egi.eu:member@cryoem.instruct-eric.eu",
                "urn:mace:egi.eu:aai.egi.eu:member@mswss.ui.savba.sk",
            ],
            "eduperson_scoped_affiliation": ["faculty@KIT"],
            "eduperson_unique_id": "d7a53cbe3e966c53ac64fde7355956560282158ecac8f3d2c770b474862f4756@egi.eu",
            "email": "hardt@kit.edu",
            "email_verified": True,
            "family_name": "Hardt",
            "given_name": "Marcus",
            "name": "Marcus Hardt",
            "preferred_username": "mhardt",
            "sub": "d7a53cbe3e966c53ac64fde7355956560282158ecac8f3d2c770b474862f4756@egi.eu",
            "iss": "https://aai.egi.eu/oidc/",
            "voperson_verified_email": ["hardt@kit.edu"],
        }
    },
}


minimal_user_info = {
    "state_target": "deployed",
    "user": {
        "userinfo": {
            "email": "hardt@kit.edu",
            "sub": "d7a53cbe3e966c53ac64fde7355956560282158ecac8f3d2c770b474862f4756@egi.eu",
            "iss": "https://aai.egi.eu/oidc/",
        }
    },
}

short_name_user_info = {
    "state_target": "deployed",
    "user": {
        "userinfo": {
            "email": "hardt@kit.edu",
            "family_name": "Ha",
            "given_name": "Marcus",
            "name": "Marcus Ha",
            "preferred_username": "mhardt",
            "sub": "d7a53cbe3e966c53ac64fde7355956560282158ecac8f3d2c770b474862f4756@egi.eu",
            "iss": "https://aai.egi.eu/oidc/",
        }
    },
}


@pytest.mark.parametrize("data", [marcus_userinfo_json])
def test_friendly_name(userinfo):
    """test friendly name generator with provided userinfo json
    Depends on the hardcoded strategies"""
    name_generator = name_generators.FriendlyNameGenerator(userinfo=userinfo)
    name = name_generator.suggest_name()
    assert name == "mhardt"

    name = name_generator.suggest_name()
    assert name == "marcus"


@pytest.mark.parametrize("data", [marcus_userinfo_json])
def test_friendly_name_again(userinfo):
    """test friendly name generator with provided userinfo json
    run for a 2nd time, to verify internal state works
    Depends on the hardcoded strategies"""
    name_generator = name_generators.FriendlyNameGenerator(userinfo)
    name = name_generator.suggest_name()
    assert name == "mhardt"

    name = name_generator.suggest_name()
    assert name == "marcus"


def test_pooled_name_one():
    """test pooled name generator with provided userinfo json
    this depends on the shipped configfile
    """
    name_generator = name_generators.PooledNameGenerator("pytest")
    name = name_generator.suggest_name()
    assert name == "pytest001"

    name = name_generator.suggest_name()
    assert name == "pytest002"


def test_pooled_name():
    """test pooled name generator with provided userinfo json
    this depends on the shipped configfile
    """
    name_generator = name_generators.PooledNameGenerator(pool_prefix="pytest")
    name = name_generator.suggest_name()
    assert name == "pytest001"

    name = name_generator.suggest_name()
    assert name == "pytest002"


@pytest.mark.parametrize("data", [marcus_userinfo_json])
def test_new_friendly(userinfo):
    """test new invocation method"""
    name_generator = name_generators.NameGenerator("friendly", userinfo=userinfo)
    name = name_generator.suggest_name()
    assert name == "mhardt"

    name = name_generator.suggest_name()
    assert name == "marcus"


def test_new_pooled():
    """test new invocation method"""
    name_generator = name_generators.NameGenerator("pooled", pool_prefix="pytest")
    name = name_generator.suggest_name()
    assert name == "pytest001"

    name = name_generator.suggest_name()
    assert name == "pytest002"


@pytest.mark.parametrize("data", [marcus_userinfo_json])
def test_new_friendly_simpler(userinfo):
    """test new invocation method"""
    name_generator = name_generators.NameGenerator("friendly", userinfo=userinfo)
    name = name_generator.suggest_name()
    assert name == "mhardt"

    name = name_generator.suggest_name()
    assert name == "marcus"


def test_new_pooled_simpler():
    """test new invocation method"""
    name_generator = name_generators.NameGenerator("pooled", pool_prefix="pytest")
    name = name_generator.suggest_name()
    assert name == "pytest001"

    name = name_generator.suggest_name()
    assert name == "pytest002"


@pytest.mark.parametrize("data", [marcus_userinfo_json])
def test_new_pooled_generic(userinfo):
    """test new invocation method"""
    name_generator = name_generators.NameGenerator(
        "friendly", userinfo=userinfo, pool_prefix="pytest"
    )
    name = name_generator.suggest_name()
    assert name == "mhardt"

    name_generator = name_generators.NameGenerator(
        "pooled", userinfo=userinfo, pool_prefix="pytest"
    )
    name = name_generator.suggest_name()
    assert name == "pytest001"


@pytest.mark.parametrize(
    "data,tried,num_tries",
    [
        (
            marcus_userinfo_json,
            [
                # all strategies in order; all lowercase and replace @ with -
                "mhardt",  # f"{userinfo.preferred_username}",
                "marcus",  # f"{userinfo.given_name}",
                "marhar",  # f"{userinfo.given_name:.3}{userinfo.family_name:.3}",
                "hardt",  # f"{userinfo.family_name}",
                "marchar",  # f"{userinfo.given_name:.4}{userinfo.family_name:.3}",
                "marcuhar",  # f"{userinfo.given_name:.5}{userinfo.family_name:.3}",
                "mahar",  # f"{userinfo.given_name:.2}{userinfo.family_name:.3}",
                "marchard",  # f"{userinfo.given_name:.4}{userinfo.family_name:.4}",
                "marcuhard",  # f"{userinfo.given_name:.5}{userinfo.family_name:.4}",
                "mahard",  # f"{userinfo.given_name:.2}{userinfo.family_name:.4}",
                "marchardt",  # f"{userinfo.given_name:.4}{userinfo.family_name:.5}",
                "marcuhardt",  # f"{userinfo.given_name:.5}{userinfo.family_name:.5}",
                "mahardt",  # f"{userinfo.given_name:.2}{userinfo.family_name:.5}",
                "marcha",  # f"{userinfo.given_name:.4}{userinfo.family_name:.2}",
                "marcuha",  # f"{userinfo.given_name:.5}{userinfo.family_name:.2}",
                "maha",  # f"{userinfo.given_name:.2}{userinfo.family_name:.2}",
                "hardt-kit.edu",  # f"{userinfo.email}",
            ],
            17,
        ),
        (
            minimal_user_info,
            [
                # no preferred_username, no given_name, no family_name, only email is a valid strategy
                "hardt-kit.edu"
            ],
            1,
        ),
        (
            short_name_user_info,
            [
                # truncation when length is shorter than desired length results in full name being used
                # duplicates (names already tried) are ignored
                "mhardt",  # f"{userinfo.preferred_username}",
                "marcus",  # f"{userinfo.given_name}",
                "marha",  # f"{userinfo.given_name:.3}{userinfo.family_name:.3}",
                "ha",  # f"{userinfo.family_name}",
                "marcha",  # f"{userinfo.given_name:.4}{userinfo.family_name:.3}",
                "marcuha",  # f"{userinfo.given_name:.5}{userinfo.family_name:.3}",
                "maha",  # f"{userinfo.given_name:.2}{userinfo.family_name:.3}",
                # "marcha", # f"{userinfo.given_name:.4}{userinfo.family_name:.4}",
                # "marcuha", # f"{userinfo.given_name:.5}{userinfo.family_name:.4}",
                # "maha", # f"{userinfo.given_name:.2}{userinfo.family_name:.4}",
                # "marcha", # f"{userinfo.given_name:.4}{userinfo.family_name:.5}",
                # "marcuha", # f"{userinfo.given_name:.5}{userinfo.family_name:.5}",
                # "maha", # f"{userinfo.given_name:.2}{userinfo.family_name:.5}",
                # "marcha", # f"{userinfo.given_name:.4}{userinfo.family_name:.2}",
                # "marcuha", # f"{userinfo.given_name:.5}{userinfo.family_name:.2}",
                # "maha", # f"{userinfo.given_name:.2}{userinfo.family_name:.2}",
                "hardt-kit.edu",  # f"{userinfo.email}",
            ],
            8,
        ),
    ],
)
def test_multiple_calls_friendly(userinfo, tried, num_tries):
    """test returning the tried names"""
    name_generator = name_generators.NameGenerator(
        "friendly", userinfo=userinfo, pool_prefix="pytest"
    )
    old_name = ""
    for _ in range(0, num_tries):
        new_name = name_generator.suggest_name()
        assert old_name != new_name
        old_name = new_name
    tried_names = name_generator.tried_names()
    assert isinstance(tried_names, list)
    assert len(tried_names) == num_tries
    assert tried_names == tried
    # try again, no strategy left
    assert name_generator.suggest_name() is None


@mock.patch("ldf_adapter.name_generators.CONFIG.username_generator.pool_digits", 1)
@pytest.mark.parametrize("data", [marcus_userinfo_json])
def test_multiple_calls_pooled(userinfo):
    """test returning the tried names"""
    name_generator = name_generators.NameGenerator(
        "pooled", userinfo=userinfo, pool_prefix="pytest"
    )
    old_name = ""
    for i in range(1, 10):
        new_name = name_generator.suggest_name()
        assert old_name != new_name
        assert new_name == f"pytest{i}"
        old_name = new_name
    tried_names = name_generator.tried_names()
    assert isinstance(tried_names, list)
    assert tried_names == []
    # try again, no name left
    assert name_generator.suggest_name() is None


@pytest.mark.parametrize("data", [marcus_userinfo_json])
def test_tried_names_friendly(userinfo):
    """test returning the tried names"""
    name_generator = name_generators.NameGenerator(
        "friendly", userinfo=userinfo, pool_prefix="pytest"
    )
    tried_names = name_generator.tried_names()
    assert isinstance(tried_names, list)


@pytest.mark.parametrize("data", [marcus_userinfo_json])
def test_tried_names_pooled(userinfo):
    """test returning the tried names"""
    name_generator = name_generators.NameGenerator(
        "pooled", userinfo=userinfo, pool_prefix="pytest"
    )
    tried_names = name_generator.tried_names()
    assert tried_names == []


# vim: tw=100



# vim: tw=100 foldmethod=expr
import pytest
from unittest import mock
from itertools import repeat
import logging

from ldf_adapter.userinfo import UserInfo
from . import settings

logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "sub",
    [
        "MWMQb4ybpHVSThMGpRKkqFDJIYlGLXl1CWXSRgM8bQGR9mMXRXtMbLFubL8Sua6vZn8Dq9X3YGoKR",
        "bNAkgXeaN2rlP83UeckV0fSjU2qNmKjQ7BsOsGFC7KB1PHtYGxRXkdSZ6S1egB085cwkIYt0NNPe",
        "0Y7zkfbvLmFVclSBFqtioE0xAaV4ZtMJ7tHxScwAN6FKoPn9R3aSXjoqp1jxRWFnyN7kNoq0nvJ33f7",
        "CO6bVYGIPtzPMHpm9o13S1bztxv6jHVEAXsGX3yBRmSv8RNnVsyzjJ67bvuvl4Tq2T8rPNTcfRRqgFF",
        "X5MOmSBpHzNj3h2BzHwhCFnHzMEpT7BxJEh6X0sgpqp7TBuUdLnfBIoovpQGEYZ7zxloVgmBV5B0FUa",
        "QHp9whOAuvUgmbJBUas1haLQfa4y4VnRlxbJwqm7pqvOfZudZAbDCXlIi20ifqfBYyjQpFN9XbOUSHi",
        "bQZO9gf6K08rlRilSYNlIuroaUBvlTV2QmHFHUlm8HUWZHs8RlNnvxQZXZZ62Oju3GARJQTx8Vb8vkF",
        "YqZE73sSxnOgoJ4jW0AhFfyZZPoaQGuzWZljOL4Nmt4x6VITfrH8EzJNES2uc0mOfGZrEIWUrxCHD6J",
        "LrcvjVB6q3omCRjzoX2uLJZzurDcU0srxJqoSmOSdCvdxisxvRo9nj52jCUxYCEnEW3FV4fKFt25Cqq4",
        "p32UOxqQGTh7xD8mfY1w9XVc0j2II06oXjw19iGhWUDMOMJiOFbwbwQRR6XVyHDwnR5hPA4bWlAxi6SK",
        "TRJvjHpnkwdsKpCYyjOPbbk5Zvk5YDfaO0BcRXpyntpuUQg5WLu9esrkqxHxH8o28A6An2bSG8mIP",
        "wl1dBQBFgPHFgetwz9m6VOYeTzGn2Pu3AMDtPRX6V0XvzhxhL7y9iEatlPqiSl5OO0yiSQWDz9GJ4Ut",
        "wTAg6ewnKtcpdj1t83NWojg7dr7ydjMhZf93hLCR5NA4GtlVfAJGwmy5XGO0AVMWMO0jhAw2KMhVK5t",
        "IkYQVi17paCZdkURaFput1p5FA2rCemT8szPIwtlDnbeecVuzEdsYVT7C3zXL8TTPKQXDdMxKYbibWO5",
        "yp1Xe4fhSGwdEyqzeejRwL8wxVeiMmamjcydJmR5hjIbQ5AQoN7pe4PwGSvNKXvkQmUO09ju0gVFALh",
        "QP8b8AwjfCw7n20WAyOYDCLztANZaKvc4uCyVtYwJnFWHYaXqFLlolPc9LJ7bb7shaRvFDpQHt07Qo",
        "nScpPJSkxQja7eMtgjM83MILbYarScUR4PWlLXK40rLHavl4dDty2OV6QJLH8AVs7LMtaOrWieHAiY",
        "UuTbVNi2RknFwKfy2n3XfCqkrscrIi0cdpJXVrbzIxBqgcvOOCWGL9YFXVTjWhbFjCvzxFgTR22yKw0X",
        "Vb15x73G3eYqtCEaqjM2MaGtUmKxNansoJxPhff5wtIK1VhDO6PutZrYxEAsxoZYIrhIZMrZpc926h",
        "w72mv9VQ62SDcBGFw4Izw7Vj5eH0JYRXdN0xImmIfZtAszqFz5mrB5aBxTZKqieTExDJWlrlAGfHIAZ",
    ],
)
def test_sub_masked_for_bwidm_eppn_unchanged(sub):
    info = UserInfo({"user": {"userinfo": {"sub": sub}}})
    assert info._sub_masked_for_bwidm_eppn() == sub


# Generated with: cat /dev/urandom | tr -dc '[:graph:]' | tr -d 'a-zA-Z0-9_!#$%&*+/=?{|}~^.\-' | fold -w 80 | head -n 20 | sed 's/\\/\\\\/g;s/'\''/\\'\''/g;s/^/'\''/;s/$/'\'',/'
@pytest.mark.parametrize(
    "sub",
    [
        ')@\\\'>""\'`,,();@"",,[];>,@,((::;;\'\']<)[:]"@;<\\<",])]>\'][@`<[\\<)>();("`"@\';;[(;[@;',
        ',`@:(,(>>;,()],["(@,\'"<]:[[,::;,@>`<[;\\:,@,(",:;)]`::;;>@<](><:(],@]\\:\\\'\'>>:\\\\>"',
        '@;<@,)]`[],:,,;\\>,](":,]@"@\'>)>]]@(]```(\\(),"(\\[,"(<`<[):;>`(:`,>,\':<[>`";>),(@@',
        '<`:;>``,@";))[\']@((>:)>`@\'":)>:"`,)><,\\""`,\';)<`\\>]`\'((\\>("<]\\\\\'>\\>\\\\"">],\\[<\'[\\',
        "''\\\\'>><'>]'`'<)>,(;:,`)'>'];,,]@@\";]]',[\\([;`([>@[\\:;)@(<[@\"`,`@''\"));,\\><`>>,@",
        "(`>@]`)':]>@\"\"@@<:;::,>\\\\(\":`<`]'(<@')[\\'('@\\:\";\"`(<[:][[@<,\"),<\\,;));'>`'\\>':[)",
        "(]`\\\"\\`'(;;)\\[\\)`\\[`)](>`]`)\\[]\\:>::@;\\[',[><`:\">@@,(@'`[@@````,`<,'])(\\[]'(]<[<",
        ']<[\'`)<`[)>\\<]\']@>,(""["<,":](""\'>,,\'@\\]><)\\\',(\\"("[``\'@];<"(:>[>\')\\;)::,[:;>",[',
        "\"(()''\\'@)]:;:;'<\":])`\\(`'(](),@;:]'>);\";'\\((::[;><)'(<[`(::(>>(@)(;\\,@[`]):,`,)",
        "`)<\"']@:@):\"]`\\);:\\\\`\">)\\@>><;>)],<:;\\],`<[]])('\\\\,(,\\:`(@>;'[))<,,\"];@[`,<>\\\\';",
        '(>`";>\'\'@;)")\'\\;<\\`<:<,"]@:(:)">\\@\\<[;)[:<@"\\(],(":)<\'"<,<`>:)@\';@`]:,[\\\\@<[\'>>\\',
        "])``\\;\"]`):\\`),):';';['\")>,`[<\">\":>>('](,[)<'@;:[[@@]`@;((@<<;`\\():;[:<,'`>>\\[\",",
        "]]\"<@``;;<<]);<];):'[,<<>\\>))>(>`)\")'(`<]>:@<:;,@>)';(:)>))'';::>[]<]'`)@[\"<`<](",
        ')<>][`";["",():`]@@[`];(;)\'\\\':,]`[\\[@]\'"";":<",[)())\\<,];"<,\'"<,:]"::(]<>@](@)<\\',
        '<;::>\'];(>[;;],]:(@<<:",<\'>\\@,""`(@\\\\@)@\'<[`,:<(\'`;<@\\"><@>:[](,`<]""`@"[""[)(\\:',
        "::>\"(:(())[\\'(,['((`):']';`\"]`@,';:];@<`<:\\,;:\">:)<'@(()]\\\"<;\"(>[());[@@:][:;,]`",
        ';``@[,,[@<;;,]<[";\'`["":;<[(;:])[;@\'\'>@<():`)""<"<\\,@><@@)\']<@"@`]\'(@>(<@@@\'`\'":',
        "@'[<,[>:@@[`(;:[<<](:@)<<,>\">\"[\\,:'@']';,\\\\\"(]:<,`<\",>],'>`:[,)(@([:>\\\\@]):@\\;,>",
        '))"]["(@,[[\']]<);)\'<)`@,@:"`\'`">"<[\'])\';;("]:,"\\(]@>@\'`;:()(`>)>[]>)<<,>:;,:;@[;',
        ':)]](,);\\\'(\\,,:\\`(<`<\'[@(,;[;<""""`<<(,@@",)`><@[)\'`"`]<]<\\\\:(`\\`>`<;`;(["\\,[[[]',
    ],
)
def test_sub_masked_for_bwidm_eppn_rnd(sub):
    info = UserInfo({"user": {"userinfo": {"sub": sub}}})
    assert info._sub_masked_for_bwidm_eppn() == "".join(repeat("-", 80))


@pytest.mark.parametrize(
    "iss", ["example.org", "http://example.org", "https://example.org"]
)
def test_iss_masked_for_bwidm_eppn_fixes_prefix(iss):
    info = UserInfo({"user": {"userinfo": {"iss": iss}}})
    assert info._iss_masked_for_bwidm_eppn() == "example.org"


@pytest.mark.parametrize(
    "raw,cooked",
    [
        ("exämple.org", "example.org"),
        ("example.örg", "example.org"),
        ("ürsula.org", "ursula.org"),
    ],
)
def test_iss_masked_for_bwidm_eppn_fixes_umlaute(raw, cooked):
    info = UserInfo({"user": {"userinfo": {"iss": raw}}})
    assert info._iss_masked_for_bwidm_eppn() == cooked


@pytest.mark.parametrize(
    "raw,cooked",
    [
        ("example.org/foobar", "example.org-foobar"),
        ("example.org/foo%20bar", "example.org-foo-20bar"),
    ],
)
def test_iss_masked_for_bwidm_eppn_fixes_urls(raw, cooked):
    info = UserInfo({"user": {"userinfo": {"iss": raw}}})
    assert info._iss_masked_for_bwidm_eppn() == cooked


@pytest.mark.parametrize(
    "raw,cooked",
    [
        ("fooBarBaz", "foo_bar_baz"),
        ("FooBarBaz", "foo_bar_baz"),
        ("kit-edu_LSDF-DIS", "kit-edu_lsdf-dis"),
        ("kit-edu_bwGrid", "kit-edu_bw_grid"),
        ("kit-edu_bwLSDF-FS", "kit-edu_bw_lsdf-fs"),
        ("kit-edu_bwUniCluster", "kit-edu_bw_uni_cluster"),
    ],
)
def test_group_masked_for_bwidm_converts_camel_to_snake_case(raw, cooked):
    info = UserInfo({"user": {"userinfo": {}}})
    assert info._group_masked_for_bwidm(raw) == cooked


@pytest.mark.parametrize(
    "raw,cooked",
    [
        ("FOOBARBAZ", "foobarbaz"),
        ("FOO-BAR-BAZ", "foo-bar-baz"),
    ],
)
def test_group_masked_for_bwidm_all_caps(raw, cooked):
    info = UserInfo({"user": {"userinfo": {}}})
    assert info._group_masked_for_bwidm(raw) == cooked


@pytest.mark.parametrize(
    "raw,cooked",
    [
        ("42", "four_2"),
        ("--test--", "test--"),
        ("__init__()", "init__--"),
        ("?!#_bullshit", "bullshit"),
    ],
)
def test_group_masked_for_bwidm_fixes_beginning(raw, cooked):
    info = UserInfo({"user": {"userinfo": {}}})
    assert info._group_masked_for_bwidm(raw) == cooked


@pytest.mark.parametrize("data", settings.ALL_INPUT)
def test_given_name(userinfo):
    assert userinfo.given_name == "Marcus"


@pytest.mark.parametrize("data", settings.ALL_INPUT)
def test_family_name(userinfo):
    assert userinfo.family_name == "Hardt"


@pytest.mark.parametrize("data", settings.ALL_INPUT)
def test_full_name(userinfo):
    assert userinfo.full_name == "Marcus Hardt"


@pytest.mark.parametrize(
    "data,username",
    [
        (settings.INPUT_UNITY, "marcus"),
        (settings.INPUT_EGI, "mhardt"),
        (settings.INPUT_DEEP_IAM, "marcus"),
        (settings.INPUT_INDIGO_IAM, "marcus"),
        (settings.INPUT_KIT, "lo0018"),
    ],
)
def test_username(userinfo, username):
    assert userinfo.username == username


@pytest.mark.parametrize(
    "data,email",
    [
        (settings.INPUT_UNITY, "marcus.hardt@kit.edu"),
        (settings.INPUT_EGI, "marcus.hardt@kit.edu"),
        (settings.INPUT_DEEP_IAM, None),
        (settings.INPUT_INDIGO_IAM, None),
        (settings.INPUT_KIT, "marcus.hardt@kit.edu"),
    ],
)
def test_email(userinfo, email):
    assert userinfo.email == email


@pytest.mark.parametrize(
    "data,unique_id",
    [
        (
            settings.INPUT_UNITY,
            "6c611e2a-2c1c-487f-9948-c058a36c8f0e@https%3A%2F%2Flogin.helmholtz-data-federation.de%2Foauth2",
        ),
        (
            settings.INPUT_EGI,
            "d7a53cbe3e966c53ac64fde7355956560282158ecac8f3d2c770b474862f4756%40egi.eu@https%3A%2F%2Faai.egi.eu%2Foidc%2F",
        ),
        (
            settings.INPUT_DEEP_IAM,
            "d9730f60-3b19-4f45-83ab-f29addf72d58@https%3A%2F%2Fiam.deep-hybrid-datacloud.eu%2F",
        ),
        (
            settings.INPUT_INDIGO_IAM,
            "a1ea3aa2-8daf-41bb-b4fb-eb88f439e446@https%3A%2F%2Fiam-test.indigo-datacloud.eu%2F",
        ),
        (
            settings.INPUT_KIT,
            "4cbcd471-1f51-4e54-97b8-2dd5177e25ec@https%3A%2F%2Foidc.scc.kit.edu%2Fauth%2Frealms%2Fkit%2F",
        ),
    ],
)
def test_unique_id(userinfo, unique_id):
    assert userinfo.unique_id == unique_id


@pytest.mark.parametrize(
    "data,eppn",
    [
        (
            settings.INPUT_UNITY,
            "6c611e2a-2c1c-487f-9948-c058a36c8f0e@login.helmholtz-data-federation.de-oauth2",
        ),
        (
            settings.INPUT_EGI,
            "d7a53cbe3e966c53ac64fde7355956560282158ecac8f3d2c770b474862f4756-egi.eu@aai.egi.eu-oidc-",
        ),
        (
            settings.INPUT_DEEP_IAM,
            "d9730f60-3b19-4f45-83ab-f29addf72d58@iam.deep-hybrid-datacloud.eu-",
        ),
        (
            settings.INPUT_INDIGO_IAM,
            "a1ea3aa2-8daf-41bb-b4fb-eb88f439e446@iam-test.indigo-datacloud.eu-",
        ),
        (
            settings.INPUT_KIT,
            "4cbcd471-1f51-4e54-97b8-2dd5177e25ec@oidc.scc.kit.edu-auth-realms-kit-",
        ),
    ],
)
def test_eppn(userinfo, eppn):
    assert userinfo.eppn == eppn


@pytest.mark.parametrize(
    "data,groups",
    [
        (
            settings.INPUT_UNITY,
            [
                "h-df-de_imk-tro-ewcc",
                "h-df-de_my_example_colab",
                "h-df-de_wlcg-test",
                "h-df-de_hdf",
            ],
        ),
        (settings.INPUT_EGI, []),
        (
            settings.INPUT_EGI_MANYGROUPS,
            [
                "egi-eu_covid19-eosc-synergy-eu_admins",
                "egi-eu_covid19-eosc-synergy-eu_admins",
                "egi-eu_cryoem-instruct-eric-eu_admins",
                "egi-eu_cryoem-instruct-eric-eu_admins",
                "egi-eu_eosc-synergy-eu_admins",
                "egi-eu_eosc-synergy-eu_admins",
                "egi-eu_eosc-synergy-eu",
                "egi-eu_eosc-synergy-eu",
                "egi-eu_goc-egi-eu",
                "egi-eu_goc-egi-eu",
                "egi-eu_mteam-data-kit-edu_admins",
                "egi-eu_mteam-data-kit-edu_admins",
                "egi-eu_mteam-data-kit-edu_perfmon-m-d-k-e_admins",
                "egi-eu_mteam-data-kit-edu_perfmon-m-d-k-e",
                "egi-eu_mteam-data-kit-edu",
                "egi-eu_mteam-data-kit-edu",
                "egi-eu_o3as-data-kit-edu_admins",
                "egi-eu_o3as-data-kit-edu_admins",
                "egi-eu_o3as-data-kit-edu",
                "egi-eu_o3as-data-kit-edu",
                "egi-eu_registry_perfmon",
                "egi-eu_registry_perfmon",
                "egi-eu_saps-vo-i3m-upv-es_admins",
                "egi-eu_saps-vo-i3m-upv-es_admins",
                "egi-eu_umsa-cerit-sc-cz_admins",
                "egi-eu_umsa-cerit-sc-cz_admins",
                "egi-eu_university-eosc-synergy-eu_admins",
                "egi-eu_university-eosc-synergy-eu_admins",
                "egi-eu_university-eosc-synergy-eu",
                "egi-eu_university-eosc-synergy-eu",
                "egi-eu_worsica-vo-incd-pt",
                "egi-eu_worsica-vo-incd-pt",
            ],
        ),
        (settings.INPUT_DEEP_IAM, ["kit-cloud"]),
        (settings.INPUT_INDIGO_IAM, ["users", "developers", "test-vo-users"]),
        (
            settings.INPUT_KIT,
            [
                "kit-edu_dfn-slcs",
                "kit-edu_lsdf-dis",
                "kit-edu_bw_grid",
                "kit-edu_bw_lsdf-fs",
                "kit-edu_bw_uni_cluster",
                "kit-edu_bwsyncnshare",
                "kit-edu_bwsyncnshare-idm",
                "kit-edu_gruppenverwalter",
            ],
        ),
    ],
)
def test_groups_classic(userinfo, groups):
    assert sorted(userinfo.groups) == sorted(set(groups))


@pytest.mark.parametrize(
    "data,groups",
    [
        (
            settings.INPUT_UNITY,
            [
                "h-df-de_imk-tro-ewcc",
                "h-df-de_my_example_colab",
                "h-df-de_wlcg-test",
                "h-df-de_hdf",
            ],
        ),
        (settings.INPUT_EGI, []),
        (
            settings.INPUT_EGI_MANYGROUPS,
            [
                "egi-eu_covid19-eosc-synergy-eu_admins",
                "egi-eu_covid19-eosc-synergy-eu_admins",
                "egi-eu_cryoem-instruct-eric-eu_admins",
                "egi-eu_cryoem-instruct-eric-eu_admins",
                "egi-eu_eosc-synergy-eu_admins",
                "egi-eu_eosc-synergy-eu_admins",
                "egi-eu_eosc-synergy-eu",
                "egi-eu_eosc-synergy-eu",
                "egi-eu_goc-egi-eu",
                "egi-eu_goc-egi-eu",
                "egi-eu_mteam-data-kit-edu_admins",
                "egi-eu_mteam-data-kit-edu_admins",
                "egi-eu_mteam-data-kit-edu_perfmon-m-d-k-e_admins",
                "egi-eu_mteam-data-kit-edu_perfmon-m-d-k-e",
                "egi-eu_mteam-data-kit-edu",
                "egi-eu_mteam-data-kit-edu",
                "egi-eu_o3as-data-kit-edu_admins",
                "egi-eu_o3as-data-kit-edu_admins",
                "egi-eu_o3as-data-kit-edu",
                "egi-eu_o3as-data-kit-edu",
                "egi-eu_registry_perfmon",
                "egi-eu_registry_perfmon",
                "egi-eu_saps-vo-i3m-upv-es_admins",
                "egi-eu_saps-vo-i3m-upv-es_admins",
                "egi-eu_umsa-cerit-sc-cz_admins",
                "egi-eu_umsa-cerit-sc-cz_admins",
                "egi-eu_university-eosc-synergy-eu_admins",
                "egi-eu_university-eosc-synergy-eu",
                "egi-eu_university-eosc-synergy-eu",
                "egi-eu_worsica-vo-incd-pt",
                "egi-eu_worsica-vo-incd-pt",
            ],
        ),
        (settings.INPUT_DEEP_IAM, ["kit-cloud"]),
        (settings.INPUT_INDIGO_IAM, ["users", "developers", "test-vo-users"]),
        (
            settings.INPUT_KIT,
            [
                "kit-edu_dfn-slcs",
                "kit-edu_lsdf-dis",
                "kit-edu_bw_grid",
                "kit-edu_bw_lsdf-fs",
                "kit-edu_bw_uni_cluster",
                "kit-edu_bwsyncnshare",
                "kit-edu_bwsyncnshare-idm",
                "kit-edu_gruppenverwalter",
            ],
        ),
    ],
)
def test_groups_regex_1(userinfo, groups, monkeypatch):
    monkeypatch.setattr(
        "ldf_adapter.backend.local_unix.CONFIG.groups.method",
        "regex",
    )
    monkeypatch.setattr(
        "ldf_adapter.backend.local_unix.CONFIG.groups.mapping",
        r"""
        :role=(owner|member|vm_operator) -> # remove all role=member and role=owner entries
        :role= -> : # all other roles: map to :
        urn:geant:kit.edu:group: -> kit-edu_
        urn:mace:egi.eu:group: -> egi-eu_
        :perfmon -> _perfmon
        urn:geant:h-df.de:group: -> h-df-de_
        :admins -> _admins
        """,
    )
    assert sorted(set(userinfo.groups)) == sorted(set(groups))
    #  assert sorted(userinfo.groups) == sorted(groups)


@pytest.mark.parametrize(
    "data,groups",
    [
        (
            settings.INPUT_UNITY,
            [
                "this-is-a-test",
            ],
        ),
        (settings.INPUT_EGI, []),
        (
            settings.INPUT_EGI_MANYGROUPS,
            [
                "this-is-a-test",
            ],
        ),
        (
            settings.INPUT_DEEP_IAM,
            ["kit-cloud"],
        ),  # group entries are (currently) not mapped
        (settings.INPUT_INDIGO_IAM, ["developers", "test-vo-users", "users"]),
        (
            settings.INPUT_KIT,
            [
                "this-is-a-test",
            ],
        ),
    ],
)
def test_groups_regex_2(userinfo, groups, monkeypatch):
    monkeypatch.setattr(
        "ldf_adapter.backend.local_unix.CONFIG.groups.method",
        "regex",
    )
    monkeypatch.setattr(
        "ldf_adapter.backend.local_unix.CONFIG.groups.mapping",
        r"""
        ^.* -> this-is-a-test
        """,
    )
    assert sorted(set(userinfo.groups)) == sorted(set(groups))


@pytest.mark.parametrize(
    "data,groups",
    [
        (
            settings.INPUT_UNITY,
            [
                "h-df-de_my_example_colab",
            ],
        ),
        (settings.INPUT_EGI, []),
        (
            settings.INPUT_EGI_MANYGROUPS,
            [
                "egi-eu_eosc-synergy-eu_admins",
                "egi-eu_eosc-synergy-eu",
            ],
        ),
        (settings.INPUT_DEEP_IAM, ["kit-cloud"]),
        (settings.INPUT_INDIGO_IAM, ["developers", "test-vo-users"]),
        (
            settings.INPUT_KIT,
            [
                "kit-edu_bw_grid",
                "kit-edu_bw_lsdf-fs",
                "kit-edu_bw_uni_cluster",
                "kit-edu_bwsyncnshare",
                "kit-edu_bwsyncnshare-idm",
            ],
        ),
    ],
)
def test_filters(userinfo, groups, monkeypatch):
    monkeypatch.setattr(
        "ldf_adapter.backend.local_unix.CONFIG.groups.policy",
        "listed",
    )
    monkeypatch.setattr(
        "ldf_adapter.backend.local_unix.CONFIG.groups.supported_entitlements",
        r"""
        urn:mace:egi.eu:group:eosc-synergy.eu.*
        urn:geant:kit.edu:group:bw.*
        urn:geant.*MyExample.*
        """,
    )
    monkeypatch.setattr(
        "ldf_adapter.backend.local_unix.CONFIG.groups.supported_groups",
        r"""
        Developers
        KIT-Cloud
        test.vo.*
        """,
    )
    assert sorted(userinfo.groups) == sorted(groups)


@mock.patch("ldf_adapter.userinfo.CONFIG.ldf_adapter.fallback_group", "nogroup")
@mock.patch("ldf_adapter.userinfo.CONFIG.ldf_adapter.primary_group", "mytestgroup")
@pytest.mark.parametrize("data", settings.ALL_INPUT)
def test_primary_group_primary_and_fallback_configured(userinfo):
    assert userinfo.primary_group == "mytestgroup"


@pytest.mark.parametrize(
    "data,group",
    [
        (settings.INPUT_UNITY, "h-df-de_hdf"),
        #  (settings.INPUT_EGI, None),
        (settings.INPUT_EGI, "nogroup"),
        (settings.INPUT_DEEP_IAM, "kit-cloud"),
        (settings.INPUT_INDIGO_IAM, "developers"),
        (settings.INPUT_KIT, "kit-edu_bw_grid"),
    ],
)
def test_primary_group_no_fallback_or_primary_configured(userinfo, group):
    logger.warning(f"userinfo.primary_group: {userinfo.primary_group} group: {group}")
    assert userinfo.primary_group == group


@mock.patch("ldf_adapter.userinfo.CONFIG.ldf_adapter.fallback_group", "nogroup")
@pytest.mark.parametrize(
    "data,group",
    [
        (settings.INPUT_UNITY, "h-df-de_hdf"),
        (settings.INPUT_EGI, "nogroup"),
        (settings.INPUT_DEEP_IAM, "kit-cloud"),
        (settings.INPUT_INDIGO_IAM, "developers"),
        (settings.INPUT_KIT, "kit-edu_bw_grid"),
    ],
)
def test_primary_group_fallback_configured_no_primary(userinfo, group):
    assert userinfo.primary_group == group


def test_egi_sub_is_unscoped():
    assert "@" in settings.INPUT_EGI["user"]["userinfo"]["sub"]


@pytest.mark.parametrize(
    "data,assurance",
    [
        (
            settings.INPUT_UNITY,
            [
                "https://refeds.org/assurance/IAP/medium",
                "https://refeds.org/assurance/IAP/local-enterprise",
                "https://refeds.org/assurance/ID/eppn-unique-no-reassign",
                "https://refeds.org/assurance/ATP/ePA-1m",
                "https://refeds.org/assurance/ATP/ePA-1d",
                "https://refeds.org/assurance/ID/unique",
                "https://refeds.org/assurance/profile/cappuccino",
                "https://refeds.org/assurance/IAP/low",
            ],
        ),
        (settings.INPUT_EGI, ["https://aai.egi.eu/LoA#Substantial"]),
        (settings.INPUT_DEEP_IAM, []),
        (settings.INPUT_INDIGO_IAM, []),
        (settings.INPUT_KIT, []),
    ],
)
def test_missing_assurance(userinfo, assurance):
    assert sorted(userinfo.assurance) == sorted(assurance)


def test_ignore_excess_entitlement():
    """https://codebase.helmholtz.cloud/m-team/feudal/feudalAdapterLdf/issues/8"""

    input_test = {
        "eduperson_assurance": [
            "https://refeds.org/assurance/IAP/medium",
            "https://refeds.org/assurance/IAP/local-enterprise",
            "https://refeds.org/assurance/ID/eppn-unique-no-reassign",
            "https://refeds.org/assurance/ATP/ePA-1m",
            "https://refeds.org/assurance/ATP/ePA-1d",
            "https://refeds.org/assurance/ID/unique",
            "https://refeds.org/assurance/profile/cappuccino",
            "https://refeds.org/assurance/IAP/low",
        ],
        "eduperson_entitlement": [
            "urn:mace:dir:entitlement:common-lib-terms",
            "http://bwidm.de/entitlement/bwLSDF-SyncShare",
            "urn:geant:h-df.de:group:IMK-TRO-EWCC#login.helmholtz-data-federation.de",
            "urn:geant:h-df.de:group:MyExampleColab#login.helmholtz-data-federation.de",
            "urn:geant:h-df.de:group:wlcg-test#login.helmholtz-data-federation.de",
            "urn:geant:h-df.de:group:HDF#login.helmholtz-data-federation.de",
        ],
        "eduperson_scoped_affiliation": "member@kit.edu",
        "email": "marcus.hardt@kit.edu",
        "email_verified": "true",
        "family_name": "Hardt",
        "given_name": "Marcus",
        "groups": ["/wlcg-test", "/IMK-TRO-EWCC", "/MyExampleColab", "/HDF", "/"],
        "iss": "https://login.helmholtz-data-federation.de/oauth2",
        "name": "Marcus Hardt",
        "preferred_username": "marcus",
        "ssh_key": "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAIAqA5FW6m3FbFhCOsRQBxKMRki5qJxoNhZdaeLXg6ym/ marcus@nemo2019\n",
        "sub": "6c611e2a-2c1c-487f-9948-c058a36c8f0e",
    }

    info = UserInfo({"user": {"userinfo": input_test}})
    assert len(list(info.entitlement)) == 4



import pytest

from ldf_adapter.userinfo import UserInfo


@pytest.fixture(scope="function")
def userinfo(data):
    """Creates a UserInfo object from provided dict"""
    userinfo = UserInfo(data)
    yield userinfo



#!/usr/bin/env python3
# vim: tw=100 foldmethod=expr
"""Simulate how the user would be created. 
"""
import sys
import json

# Must be before the first ldf_adapter import
from feudal_globalconfig import globalconfig

from ldf_adapter import User
from ldf_adapter.cmdline_params import args


def main():
    try:
        data = json.load(sys.stdin)
    except json.decoder.JSONDecodeError as e:
        message = "Cannot decode the input json. Please verify the input!"
        print(message)
        sys.exit(2)

    try:  # User.__init__ tries finding userinfo in data["user"]["userinfo"]
        user = User(data)
    except KeyError:
        try:
            data = {"user": {"userinfo": data}}
            user = User(data)
        except KeyError:
            print("cannot find required information in input json.")
            sys.exit(3)
        except Exception as e:
            print(f"Unhandled Exception: {e}")
            raise

    print("From the input data and the provided config, feudal would make use of:")
    print(f"User: {user.data.username}")
    print("Groups:")
    for grp in sorted(user.data.groups):
        print(f"   {grp}")


if __name__ == "__main__":
    sys.exit(main())



# vim: foldmethod=indent : tw=100
# pylint: disable=invalid-name, superfluous-parens
# pylint: disable=logging-fstring-interpolation, logging-not-lazy, logging-format-interpolation
# pylint: disable=missing-docstring, too-few-public-methods
"""
BWIDM backend.

See https://git.scc.kit.edu/simon/reg-app.
"""

import logging
import json
from functools import reduce
from time import sleep
import requests
import os
from urllib.parse import urljoin

from ldf_adapter.config import CONFIG
from ldf_adapter import utils
from ldf_adapter.logsetup import jsonlogger
from ldf_adapter.backend import generic

logger = logging.getLogger(__name__)


class BwIdmConnection:
    """Connection to the BWIDM API."""

    def __init__(self, config=None):
        self.session = requests.Session()
        if config:
            self.session.auth = (
                config.backend.bwidm.http_user,
                config.backend.bwidm.http_pass,
            )

        if not CONFIG.backend.bwidm.log_outgoing_http_requests:
            logging.getLogger("requests").setLevel(logging.CRITICAL)
            logging.getLogger("werkzeug").setLevel(logging.CRITICAL)
            logging.getLogger("urllib3").setLevel(logging.CRITICAL)

    def get(self, *url_fragments, **kwargs):
        return self._request("GET", url_fragments, **kwargs)

    def post(self, *url_fragments, **kwargs):
        return self._request("POST", url_fragments, **kwargs)

    def _request(self, method, url_fragments, **kwargs):
        """
        Arguments:
        method -- HTTP Method (type: str)
        url_fragments -- The components of the URL. Each is url-encoded separately and then they are
                         joined with '/'
        fail=True -- Raise exception on non-200 HTTP status
        **kwargs -- Passed to `requests.Request.__init__`
        """
        fail = kwargs.pop("fail", True)

        url_fragments = map(str, url_fragments)
        url_fragments = map(
            lambda frag: requests.utils.quote(frag, safe=""), url_fragments
        )
        url = reduce(
            lambda acc, frag: (
                urljoin(acc, frag) if acc.endswith("/") else urljoin(acc + "/", frag)
            ),
            url_fragments,
            CONFIG.backend.bwidm.url,
        )

        # logger.debug(f"BWIDM: {url}")
        # logger.debug(f"BWIDM: {kwargs}")
        req = requests.Request(method, url, **kwargs)
        rsp = self.session.send(self.session.prepare_request(req))

        if fail:
            if not rsp.ok:
                logger.error(
                    "Server responded with: {}".format(rsp.content.decode("utf-8"))
                )
            rsp.raise_for_status()

        return rsp


BWIDM = BwIdmConnection(CONFIG)


class User(generic.User):
    ATTR_USERNAME = "urn:oid:0.9.2342.19200300.100.1.1"
    ATTR_ORG_ID = "http://bwidm.de/bwidmOrgId"
    VALUE_USER_ACTIVE = "ACTIVE"
    VALUE_USER_INACTIVE = "ON_HOLD"

    def __init__(self, userinfo, **hooks):
        super().__init__(userinfo, **hooks)
        self.info = userinfo
        self.primary_group = Group(userinfo.primary_group)
        self.force_username = None
        self.post_create_script = CONFIG.backend.bwidm.post_create_script

    def exists(self):
        """
        Inactive users ('ON_HOLD') are treated as nonexistent.
        """
        return self._exists() and self._is_active()

    def _exists(self):
        exists = b"no such user" not in self.reg_info(json=False, fail=False)
        logger.debug(
            "User {} {} on service BWIDM".format(
                self.info.unique_id, "exists" if exists else "doesn't exist"
            )
        )
        return exists

    def _is_active(self):
        status = self.reg_info()["userStatus"]
        logger.debug(
            "User {} is {} on service BWIDM".format(self.info.unique_id, status)
        )
        return status == self.VALUE_USER_ACTIVE

    def _is_registered(self):
        """
        find if the user is already registered for a given
        service, identified by its service short name
        """
        # FIXME: Consider putting this request into the global user object (to reduce load on regapp)
        ssn = CONFIG.backend.bwidm.service_name
        registrations = BWIDM.get(
            "external-reg", "find", "externalId", self.info.unique_id
        )
        # find registrations
        number_of_registrations = 0
        try:
            logger.debug("logging registrations to jsonlog")
            jsonlogger.debug(
                json.dumps(
                    registrations.json, sort_keys=True, indent=4, separators=(",", ": ")
                )
            )
        except TypeError:
            pass
        for reg in registrations.json():
            if reg["serviceShortName"] == ssn:
                if reg["registryStatus"] == "ACTIVE":
                    number_of_registrations += 1
        logger.debug(f"Number of registrations found: {number_of_registrations}")
        if number_of_registrations > 0:
            return True
        return False

    def name_taken(self, name):
        """
        If there is a user for our unique_id with our username, treat the name as available. This
        might happen if the our user is ON_HOLD on the service.
        """
        # TODO: argument "name" was added, check if given "name" (instead of info.username) is taken by *another* user
        logger.debug(f"checking for name taken of {name}")
        users_with_name = BWIDM.get(
            "external-user", "find", "attribute", self.ATTR_USERNAME, name
        ).json()

        other_users_with_name = [
            user
            for user in users_with_name
            if user["externalId"] != self.info.unique_id
        ]
        # logger.debug (F"other_users: {other_users_with_name}")
        logger.debug(f"Found {len(other_users_with_name)} with same username")
        # FIXME: Why < ?
        if len(other_users_with_name) < len(users_with_name):
            logger.debug("Username '{}' is reserved for us".format(self.info.username))

        if other_users_with_name:
            logger.error(
                "Username '{}' is already used by\n    {}".format(
                    self.info.username,
                    ",\n    ".join(
                        map(lambda u: u["externalId"], other_users_with_name)
                    ),
                )
            )
        else:
            logger.debug("Username '{}' is available".format(name))

        return bool(other_users_with_name)

    def get_username(self):
        """Check if a user exists based on unique_id and return the name"""

        def safe_resp_conversion(resp):
            """Safely convert a response to json"""
            if resp.status_code != 200:
                logger.debug(
                    "Error %d reading from remote: \n%s\n"
                    % (resp.status_code, str(resp.text))
                )
                os._exit(1)  # or raise or return None?
            try:
                resp_json = resp.json()
            except json.JSONDecodeError:
                logging.error("Could not decode json that I obtained from rest server")
                raise
            return resp_json

        username = None
        bwIdmOrgId = ""
        resp = BWIDM.get("external-user", "find", "externalId", self.info.unique_id)
        resp_json = safe_resp_conversion(resp)

        try:
            username = resp_json["attributeStore"]["urn:oid:0.9.2342.19200300.100.1.1"]
            bwIdmOrgId = resp_json["attributeStore"]["http://bwidm.de/bwidmOrgId"]
            logger.debug(f"retuning username: {bwIdmOrgId}_{username}")
            return f"{bwIdmOrgId}_{username}"
        except KeyError as e:
            logger.error("Error: I could not find the username in the database.")
            logger.error("  Most likely the user is not registered for this service\n")
            logger.error(f"  {e}")
            logger.error(
                json.dumps(resp_json, sort_keys=True, indent=4, separators=(",", ": "))
            )
        return None

    def set_username(self, username):
        """Update the internal representation of the user with the incoming username"""
        logger.debug(f"set_username: setting {username}")
        self.force_username = username

    def set_prefixed_username(self, prefixed_username):
        """Update the internal representation of the user with the incoming username"""
        bwIdmOrgId = CONFIG.backend.bwidm.org_id
        username = prefixed_username.lstrip(f"{bwIdmOrgId}_")
        self.set_username(username)

    def create(self):
        """Create or activate user."""
        if self._exists() and not self._is_active():
            logger.info("Activating user {unique_id}".format(**self.info))
            BWIDM.get("external-user", "activate", "externalId", self.info.unique_id)
        else:
            logger.info("Creating user {unique_id}".format(**self.info))
            BWIDM.post(
                "external-user", "create", json={"externalId": self.info.unique_id}
            )

    def register(self):
        """register user for the configured service"""

        def get_active_reg_info(ext_id):
            rsp = BWIDM.get("external-reg", "find", "externalId", ext_id)

            try:
                return next(
                    filter(lambda reg: reg["registryStatus"] == "ACTIVE", rsp.json())
                )
            except StopIteration:
                return {"lastReconcile": None}

        old_reg = get_active_reg_info(self.info.unique_id)
        # We wait until the 'lastReconciled' timestamp changes, which means that our update was sucessfully deployed
        reg = old_reg
        while reg["lastReconcile"] == old_reg["lastReconcile"]:
            sleep(0.3)
            logger.debug(
                "Received registration reconciled at {}. That is not up-to-date. Checking again.".format(
                    reg["lastReconcile"]
                )
            )

            rsp = BWIDM.get(
                "external-reg",
                "register",
                "externalId",
                self.info.unique_id,
                "ssn",
                CONFIG.backend.bwidm.service_name,
            )

            if rsp.status_code == 204:
                reg = get_active_reg_info(self.info.unique_id)
            else:
                reg = rsp.json()

        logger.debug(
            "Registration confirmed reconciled at {}. Looks like the update went through.".format(
                reg["lastReconcile"]
            )
        )
        logger.debug(f"user is now registered: {self._is_registered()}")

    def update(self):
        # FIXME: Consider putting this request into the global user object (to reduce load on regapp)

        self.external_user_update(
            {
                "externalId": self.info.unique_id,
                "eppn": self.info.eppn,
                "email": self.info.email,
                "givenName": self.info.given_name,
                "surName": self.info.family_name,
                "primaryGroup": {"id": self.primary_group.reg_info()["id"]},
                "attributeStore": {
                    self.ATTR_USERNAME: self.force_username or self.info.username,
                    self.ATTR_ORG_ID: CONFIG.backend.bwidm.org_id,
                },
            }
        )
        # if the user is already registered for the service, we're done here
        if not self._is_registered():
            self.register()

        logger.debug(f"user is active: {self._is_active()}")
        logger.debug(f"user is registered: {self._is_registered()}")

    def delete(self):
        """Deregister the user from the given service in BWIDM."""
        BWIDM.get(
            "external-reg",
            "deregister",
            "externalId",
            self.info.unique_id,
            "ssn",
            CONFIG.backend.bwidm.service_name,
        )

    def deactivate(self):
        """Deactivate the user, this does not delete from BWIDM, but sets
        the status to ON_HOLD, thereby disabling the user from ALL
        services."""
        BWIDM.get("external-user", "deactivate", "externalId", self.info.unique_id)

    def mod(self, supplementary_groups=None):
        reg_info = self.reg_info()

        if supplementary_groups is not None:
            current_groups = [grp for grp in reg_info["secondaryGroups"]]
            new_groups = [grp.reg_info(short=True) for grp in supplementary_groups]
            new_groups += [self.primary_group.reg_info()]

            # NL = "\n    "
            # logger.debug(
            #     f"Incoming groups: {NL}{NL.join([g['name'] for g in new_groups])}"
            # )

            # Remove user from groups he should not be a member of
            to_be_removed_from = [
                g
                for g in current_groups
                if g["id"] not in (ng["id"] for ng in new_groups)
            ]

            # Only add user to groups she is not already a member of
            to_be_added_to = [
                g
                for g in new_groups
                if g["id"] not in (cg["id"] for cg in current_groups)
            ]

            if to_be_removed_from:
                logger.info(
                    "Remove user {} from groups {}".format(
                        self.info.username,
                        ",".join(g["name"] for g in to_be_removed_from),
                    )
                )
            for grp in to_be_removed_from:
                BWIDM.get(
                    "group-admin",
                    "remove",
                    "groupId",
                    grp["id"],
                    "userId",
                    reg_info["id"],
                )

            if to_be_added_to:
                logger.info(
                    "Add user {} to groups {}".format(
                        self.info.username, ",".join(g["name"] for g in to_be_added_to)
                    )
                )
            grp_add_retvals = []
            for grp in to_be_added_to:
                grp_add_retvals.append(
                    BWIDM.get(
                        "group-admin",
                        "add",
                        "groupId",
                        grp["id"],
                        "userId",
                        reg_info["id"],
                        fail=False,
                    )
                )
            if len(grp_add_retvals) > 0:
                logger.debug(f"Group add retvals: {grp_add_retvals}")
            return to_be_added_to, to_be_removed_from
        return [], []

    def install_ssh_keys(self):
        self.external_user_update(
            {
                "externalId": self.info.unique_id,
                "genericStore": {
                    **self.reg_info()["genericStore"],
                    "ssh_key": json.dumps(self.info.ssh_keys),
                },
            }
        )

    def uninstall_ssh_keys(self):
        """Uninstall any SSH keys stored in BWIDM."""
        self.external_user_update(
            {"externalId": self.info.unique_id, "genericStore": {"ssh_key": None}}
        )

    def external_user_update(self, state_updates):
        """Apply new attributes to the user, performing sensible merging of dicts.

        BWIDM is a bit weird about this, due to technical restrictions in the Java-Software stack.

        This comes down to:
        {..., k: val, ...} means `state[k] = val`
        {..., k: None, ...} means `del state[k]` or `state[k]=None`
        {..., k: {}, ...} means no change to k
        {..., k: val={...}, ...} means `state[k]=merge state[k] with val`

        This is applied recursivly.

        """
        current_state = self.reg_info()
        new_state = utils.dictmerge(current_state, state_updates)
        utils.log_dictdiff(
            utils.dictdiff(current_state, new_state), log_function=logger.info
        )
        try:
            logger.debug(
                f"current_state: {current_state['attributeStore']['urn:oid:0.9.2342.19200300.100.1.1']}"
            )

            logger.debug(
                f"new_state:     {new_state['attributeStore']['urn:oid:0.9.2342.19200300.100.1.1']}    "
            )
            logger.debug(
                f"state_updates: {state_updates['attributeStore']['urn:oid:0.9.2342.19200300.100.1.1']}"
            )

        except KeyError:
            pass

        for k in list(new_state):
            if new_state[k] is None:
                new_state[k] = {}

        try:
            formatted_json = json.dumps(
                state_updates, sort_keys=True, indent=4, separators=(",", ": ")
            )
            logger.debug("  logging state_updates to jsonlog")
            jsonlogger.debug(f"state_updates:  {formatted_json}")

            formatted_json = json.dumps(
                current_state, sort_keys=True, indent=4, separators=(",", ": ")
            )
            logger.debug("  logging current_state_updates to jsonlog")
            jsonlogger.debug(f"current_state: {formatted_json}")

            logger.debug("  logging new_to jsonlog")
            formatted_json = json.dumps(
                new_state, sort_keys=True, indent=4, separators=(",", ": ")
            )
            jsonlogger.debug(f"new state for regapp:  {formatted_json}")
        except:
            pass
        BWIDM.post("external-user", "update", json=new_state)

    def reg_info(self, json=True, **kwargs):
        # FIXME: Cache this functions results!
        rsp = BWIDM.get(
            "external-user", "find", "externalId", self.info.unique_id, **kwargs
        )
        return rsp.json() if json else rsp.content


class Group(generic.Group):
    def __init__(self, name):
        self.name = name

    def exists(self):
        # FIXME: Group existence needs to be checked with using also
        # CONFIG.backend.bwidm.service_name
        return (
            b"no such group"
            not in BWIDM.get(
                "group-admin", "find", "name", self.name, fail=False
            ).content
        )

    def create(self):
        rsp = BWIDM.get(
            "group-admin", "create", CONFIG.backend.bwidm.service_name, self.name
        ).json()

        if self.name != rsp["name"]:
            logger.warning(
                "Groupname changed from {} to {} by BWIDM".format(
                    self.name, rsp["name"]
                )
            )
            self.name = rsp["name"]

        self.id = rsp["id"]

    def delete(self):
        # groupdel
        raise NotImplementedError("Do we even need this function?")

    def mod(self):
        # groupmod
        raise NotImplementedError("Do we even need this function?")

    def reg_info(self, json=True, short=False, **kwargs):
        rsp = BWIDM.get(
            "group-admin",
            "find" if short else "find-detail",
            "name",
            self.name,
            **kwargs,
        )
        return rsp.json() if json else rsp.content

    @property
    def members(self):
        raise NotImplementedError("Do we even need this function?")



from os.path import dirname

from ldf_adapter.config import CONFIG
from ldf_adapter.utils import create_factory


users = create_factory(
    dir_name=dirname(__file__), parent_module=__name__, class_name="User"
)
groups = create_factory(
    dir_name=dirname(__file__), parent_module=__name__, class_name="Group"
)

User = users.get_type(CONFIG.ldf_adapter.backend)
Group = groups.get_type(CONFIG.ldf_adapter.backend)



# vim: tw=100 foldmethod=expr
"""Information about the user. 
    This serves as a wrapper around the plain userinfo-dict
"""
from collections.abc import Mapping
from functools import lru_cache
from itertools import chain
import urllib.parse
from unidecode import unidecode
import logging
import regex

from ldf_adapter import eduperson
from ldf_adapter.config import CONFIG
from ldf_adapter.results import raise_question

logger = logging.getLogger(__name__)


class UserInfo(Mapping):
    """Information about the user.

    This serves as a wrapper around the plain userinfo-dict passed to us by FEUDAL, exposing only
    the required information. Provides reconstruction of attributes in case of missing information
    in the userinfo-dict (if possible), homogenisation of the values by mapping them (non-
    bijectively!) to reduced character ranges, without lobotomizing the original input to much, as
    this risks collisions.

    E.g., everything returned by this is compatible with BWIDM, but not necessarily with UNIX
    shadow-utils(7), as the latter has very strict requiremnts which probably one does not want to
    apply to all services. This, if your backend has stricter requirements, you need to perform
    further homogenisation on your own.

    Any change made to values is logged with level WARNING.

    The values are exposed as properties, calculated lazily only when needed (they are cached
    however).  Any instance can also be used as a dict, i.e. `userinfo.foo == userinfo['foo']`.

    All properties (when called as a function) take an optional boolean `allow_question`, indicating
    whether it should be allowed to raise a `Questionaire` if needed.
    """

    def __init__(self, data):
        """
        Arguments:
        data -- Input as recieved by FEUDAL
        """
        self.userinfo = data["user"]["userinfo"]
        self.answers = data.get("answers", {})
        self.credentials = data["user"].get("credentials", {})
        self.allow_question = CONFIG.ldf_adapter.interactive

    @property
    def sub(self):
        return self.userinfo["sub"]

    @property
    def iss(self):
        return self.userinfo["iss"]

    @property
    @lru_cache(maxsize=None)
    def size(self):
        the_size = 0
        for x in self.userinfo.keys():
            the_size += 1
        return the_size

    @property
    @lru_cache(maxsize=None)
    def unique_id(self):
        """Globally and uniquely identifies the user.

        This can be easily used to find out the identity of the user in the data source.

        Percent-Encodes subject and issuer and concatenates them with an '@'

        """
        return "{sub}@{iss}".format(
            sub=self._sub_masked_for_bwidm_extid(),
            iss=self._iss_masked_for_bwidm_extid(),
        )

    def _sub_masked_for_bwidm_extid(self):
        return urllib.parse.quote_plus(self.userinfo["sub"])

    def _iss_masked_for_bwidm_extid(self):
        return urllib.parse.quote_plus(self.userinfo["iss"])

    @property
    @lru_cache(maxsize=None)
    def eppn(self):
        """Uniquely identifies the user.

        At least almost. Due to homogenisations, there might be collisions. E.g. the following users
        are all indistinguishable:

        klammer(affe)@https://example.org/oauth-2
        klammer(affe)@https://example.org/oauth/2
        klammer-affe-@https://example.org/oauth-2
        klammer(affe)@http://example.org-oauth-2
        klammer-affe-@example.org-oauth-2
        """
        return "{sub}@{iss}".format(
            sub=self._sub_masked_for_bwidm_eppn(), iss=self._iss_masked_for_bwidm_eppn()
        )

    def _sub_masked_for_bwidm_eppn(self):
        """Replace invalid characters with a dash ('-').

        Usually subjects are only numbers and ascii-chars separeted by dashes, so this should not be
        much of a problem.
        """
        # FIXME: Changing the sub of a user is potentially terrible
        sub = self.userinfo["sub"]

        sub = regex.sub("[^a-zA-Z0-9_!#$%&*+/=?{|}~^.-]", "-", sub)

        if sub != self.userinfo["sub"]:
            logger.warning(
                "sub '{}' changed to '{}' to avoid confusion".format(
                    self.userinfo["sub"], sub
                )
            )
        return sub

    def _iss_masked_for_bwidm_eppn(self):
        """Strip URI-scheme, transliterate to ASCII and replace invalid characters with a dash ('-').

        Usually there is only one issuer per FQDN (which is mostly left untouched, apart from
        transliteration, since most FQDNS consist only of alphanumerics, dashes and dots), so this
        should not be much of a problem.
        """
        stripped_iss = regex.sub("^https?://", "", self.userinfo["iss"])
        iss = unidecode(stripped_iss)
        iss = regex.sub("[^a-zA-Z0-9.-]", "-", iss)

        # We don't consider stripping the http[s]-prefix a change, since we always do that anyway,
        # and there shouldn't be two different issuers `http://example.org' and `https://example.org'.
        if iss != stripped_iss:
            if CONFIG.messages.log_name_changes:
                logger.warning(
                    "Issuer '{}' changed to '{}' for general compatibilty".format(
                        stripped_iss, iss
                    )
                )
        return iss

    @property
    @lru_cache(maxsize=None)
    def username(self):
        """Return the user's name, this may be:
        - preferred_username if that was provided
        - if in interactive mode, we prompt the user
        - otherwise, None"""
        # FIXME: If this function is called, even when there is already a local user with an existing
        # name, this is a bug in the flow, that MUST be fixed
        #
        # Simply not having a "preferred_username" does not mean that we have to bother the user!!!
        # TODO (DG): does this still need to be fixed? it should work => test

        if self.allow_question:
            return self.value_or_ask(
                self.userinfo.get("preferred_username"),
                "username",
                "You have not set a global username preference. Please enter your preferred username.",
                self.allow_question,
            )
        return self.userinfo.get("preferred_username", None)

    @property
    @lru_cache(maxsize=None)
    def email(self):
        """Return the user's E-Mail Address."""
        try:
            return self.userinfo["email"]
        except KeyError:
            return None

    @property
    @lru_cache(maxsize=None)
    def given_name(self):
        """Return the user's given name. If none is provided, try to extract it from the full name."""
        try:
            given_name_from_name = " ".join(self.userinfo["name"].split(" ")[:-1])
        except KeyError:
            given_name_from_name = None
        return self.userinfo.get("given_name") or given_name_from_name or None

    @property
    @lru_cache(maxsize=None)
    def family_name(self):
        """Return the user's family name. If none is provided, try to extract it from the full name."""
        try:
            family_name_from_sn = self.userinfo.get("sn")
        except KeyError:
            family_name_from_name = None

        try:
            family_name_from_name = self.userinfo["name"].split(" ")[-1]
        except KeyError:
            family_name_from_name = None

        return (
            self.userinfo.get("family_name")
            or family_name_from_sn
            or family_name_from_name
        )

    @property
    @lru_cache(maxsize=None)
    def full_name(self):
        """Return the user's full name. If none is provided, try to assemple it from the first and given name."""
        return self.userinfo.get("name") or " ".join(
            filter(None, [self.given_name, self.family_name])
        )

    @property
    @lru_cache(maxsize=None)
    def ssh_keys(self):
        """Return the user's SSH keys."""
        return self.credentials.get("ssh_key", [])

    @property
    def entitlement(self):
        """Return the parsed entitlement attribute of the user. See `eduperson.Entitlement` for details."""
        attr = self.userinfo.get("eduperson_entitlement", [])
        if not isinstance(attr, list):
            attr = [attr]

        def try_entitlement(attr):
            try:
                return eduperson.Entitlement(attr)
            except ValueError:
                return None

        def filter_allowed_entitlements(attr):
            """filter string-based entitlements for
            allowed regular expressions from config"""
            for s_e in supported_entitlements:
                myregex = regex.compile(s_e)
                if myregex.search(attr):
                    return True
                    #  return attr

        supported_entitlements = regex.findall(
            r"[^\s]+.*", CONFIG.groups.supported_entitlements
        )
        if CONFIG.groups.policy == "listed":
            logger.debug("filtering entitlements")
            if supported_entitlements != []:
                logger.debug(f"supported_entitlements: {supported_entitlements}")
                attr = filter(filter_allowed_entitlements, attr)
        retval = filter(lambda x: x, map(try_entitlement, attr))
        return retval

    @property
    def group(self):
        """Return the unparsed group attribute of the user"""
        attr = self.userinfo.get("groups", [])
        if not isinstance(attr, list):
            attr = [attr]
        return attr

    @property
    @lru_cache(maxsize=None)
    def groups(self):
        """Return the homogenised names of the groups the user should be a member of."""
        group_method = CONFIG.groups.method
        logger.info(f"group method: {group_method}")
        # A shitty way to see if the entitlement is empty or not:
        if len([x for x in self.entitlement]) == 0:
            logger.debug("Using plain groups from 'groups' claim")
            grouplist = self.groups_from_grouplist()
            logger.debug(f"got grouplist: {grouplist}")
            for grp in grouplist:
                logger.error(f"got group: {grp}")
        else:
            logger.debug("Using aarc-g002 groups from 'entitlements' claim")
            if group_method == "classic":
                logger.info("method: classic")
                grouplist = self.groups_from_entitlement()
            elif group_method == "regex":
                logger.info("method: regex")
                grouplist = self.groups_from_entitlement_mapped()
            else:  # the default...
                logger.info("method: default")
                grouplist = self.groups_from_entitlement()

        return [self._group_masked_for_bwidm(grp) for grp in grouplist]

    def groups_from_entitlement_mapped(self) -> list:
        """Return a list of groups based on map in config"""
        group_list = regex.findall(r"[^\s]+.*", CONFIG.groups.mapping)
        group_map = [x.split(" -> ") for x in group_list]

        # fix missing capability of empty string:
        for map_entry in group_map:
            if len(map_entry) != 2:
                map_entry[0] = map_entry[0].rstrip(" ->")
                map_entry.append("")
        # strip comments
        for map_entry in group_map:
            if len(map_entry) >= 1:
                myregex = regex.compile(r"(^#|\W#).*")
                map_entry[0] = myregex.sub("", map_entry[0])
                map_entry[1] = myregex.sub("", map_entry[1])

        grouplist = []
        for orig_ent in self.entitlement:
            #  logger.info(F"orig_ent: {orig_ent}")
            ent = orig_ent.__repr__().split("#")[0]
            for map_entry in group_map:
                myregex = regex.compile(map_entry[0])
                ent = myregex.sub(map_entry[1], str(ent))
            logger.info(f"{orig_ent.__repr__():75} -> {ent}")
            if ent is not None:
                if len(ent) > 32:
                    logger.warning(f"Group needs shortening: {ent}")

            grouplist.append(ent)
        return grouplist

    def groups_from_entitlement(self):
        """Gropus are extracted from the entitlement.
        Any additional 'group'-keys in the input are ignored.

        Group names are prefixed with the delegated namespace from the entitlement.
        """
        if CONFIG.username_generator.strip_sub_groups:
            logger.debug("Stripping all subgroups")
            retval = set(
                filter(
                    None,
                    [
                        "{}_{}".format(ns, grp)
                        for (ns, grp) in chain.from_iterable(
                            (
                                (
                                    "-".join(
                                        [ent.delegated_namespace] + ent.subnamespaces
                                    ),
                                    grp,
                                )
                                for grp in ent.all_toplevel_groups
                            )
                            for ent in self.entitlement
                        )
                    ],
                )
            )
        else:
            retval = set(
                filter(
                    None,
                    [
                        "{}_{}".format(ns, grp)
                        for (ns, grp) in chain.from_iterable(
                            (
                                (
                                    "-".join(
                                        [ent.delegated_namespace] + ent.subnamespaces
                                    ),
                                    grp,
                                )
                                for grp in ent.all_groups
                            )
                            for ent in self.entitlement
                        )
                    ],
                )
            )
        for ent in retval:
            logger.info(f"  mapped entitlement to group: {ent}")
        return retval

    def groups_from_grouplist(self):
        """Gropus are extracted from the groups claim"""

        def filter_allowed_groups(attr):
            """filter string-based entitlements for
            allowed regular expressions from config"""
            for s_e in supported_groups:
                myregex = regex.compile(s_e)
                if myregex.search(attr):
                    return True
                    #  return attr

        supported_groups = regex.findall(r"[^\s]+.*", CONFIG.groups.supported_groups)
        groups = set([grp for grp in self.group])
        logger.info(f"groups: {groups}")

        if CONFIG.groups.policy == "listed":
            logger.debug("filtering groups")
            if supported_groups != []:
                logger.warning(f"supported_groups: {supported_groups}")
                groups = filter(filter_allowed_groups, groups)

        # groups is an iterable, which an only be expanded once
        return [x for x in groups]

    def _group_masked_for_bwidm(self, orig_grp):
        """Convert camelCase to snake_case,
        fixup beginning of name
        and replace invalid chars with a dash ('-')"""
        grp = orig_grp

        # camelCase to snake_case
        grp = regex.sub(
            "([a-z])([A-Z])",
            lambda m: "{}_{}".format(m.group(1), m.group(2).lower()),
            grp,
        )

        # Lowercase all
        grp = regex.sub("[A-Z]", lambda m: m.group(0).lower(), grp)

        # Catch remaining chars
        grp = regex.sub("[^a-z0-9-_]", "-", grp)

        # First char has to be [a-z]
        grp = regex.sub("^[-_]*", "", grp)
        grp = regex.sub("^0", "zero_", grp)
        grp = regex.sub("^1", "one_", grp)
        grp = regex.sub("^2", "two_", grp)
        grp = regex.sub("^3", "three_", grp)
        grp = regex.sub("^4", "four_", grp)
        grp = regex.sub("^5", "five_", grp)
        grp = regex.sub("^6", "six_", grp)
        grp = regex.sub("^7", "seven_", grp)
        grp = regex.sub("^8", "eight_", grp)
        grp = regex.sub("^9", "nine_", grp)

        if grp != orig_grp:
            if CONFIG.messages.log_name_changes:
                logger.warning(
                    "Group name '{}' changed to '{}' for general compatibilty".format(
                        orig_grp, grp
                    )
                )
        return grp

    @property
    @lru_cache(maxsize=None)
    def assurance(self):
        """Return the assurance levels of the user."""
        return self.userinfo.get("eduperson_assurance", [])

    @property
    @lru_cache(maxsize=None)
    def primary_group(self):
        config_group = CONFIG.ldf_adapter.primary_group
        logger.debug(f"Using configured primary group: {config_group}")
        if config_group:
            return config_group
        elif len(self.groups) == 1:
            # lousy way to access a set element:
            for group in self.groups:
                return group
        elif len(self.groups) > 1:
            if self.allow_question:
                return self.value_or_ask(
                    self.userinfo.get(0),
                    "primary_group",
                    "You are a member of multiple groups. Please select your desired primary group.",
                    self.allow_question,
                    list(self.groups),
                )
            else:  # make something up, regarding the primary group:
                if CONFIG.messages.log_primary_group_definition:
                    logger.warning(
                        "/----- No primary group issue --------------------------------------------\\"
                    )
                    logger.warning(
                        f"    We have a user with mutiple primary groups, and no default primary group set in the config."
                    )
                    logger.warning(
                        f"    Furthermore, we are in non-interactive mode, so we can't ask the user."
                    )
                    logger.warning(
                        f"    Therefore, we just take the first group: '{sorted(list(self.groups))[0]}'"
                    )
                    nl = "\n                            "
                    # logger.warning(F"    Available groups are: {nl}{nl.join(self.groups)}")
                    logger.warning(
                        "\\--------------------------------------------------------------------------------/"
                    )
                return sorted(list(self.groups))[0]

        else:
            old_answer = self.answers.get("primary_group", None)
            if old_answer is not None:
                return old_answer

            else:  # still no group found.
                fallback_group = CONFIG.ldf_adapter.fallback_group
                if fallback_group:
                    return fallback_group
                else:
                    logger.warning(
                        "Not a single group found; This may be ok, depending on the request type"
                    )
            # raise Failure(message="No groups in userinfo and no global primary group configured")

    @property
    @lru_cache(maxsize=None)
    def preferred_username(self):
        """Return the prefrred username of the user."""
        return self.userinfo.get("preferred_username", None)

    def value_or_ask(
        self, value, answer_name, question_text, allow_question, default=None
    ):
        """Return the submitted answer, the default value or raise a questionaire."""
        previous_answer = self.answers.get(answer_name)
        return (
            previous_answer
            or value
            or (
                allow_question
                and raise_question(
                    name=answer_name,
                    text=question_text,
                    default=default,
                )
            )
        )

    def __str__(self):
        attrs = (
            "{} = {}".format(k, getattr(UserInfo, k).fget(self)) for k in iter(self)
        )

        return "<UserInfo\n{}\n>".format(
            "\n".join("\t{}".format(attr) for attr in attrs)
        )

    def __getitem__(self, key):
        return getattr(self, key, lambda: (_ for _ in ()).throw(KeyError(key)))

    def __iter__(self):
        return (k for k in dir(UserInfo) if type(getattr(UserInfo, k)) is property)

    def __len__(self):
        return sum(
            1
            for _ in filter(
                lambda k: type(getattr(UserInfo, k)) is property, dir(UserInfo)
            )
        )

    def __hash__(self):
        return id(self)  # Good enough for lru_cache



from dataclasses import fields
from typing import Optional, Type, Union, List, TypeVar
from dataclasses import fields
import sqlite3
import json
import logging
from pathlib import Path

from ldf_adapter.results import Failure, FatalError
from ldf_adapter.config import CONFIG
from ldf_adapter.approval.models import (
    DeploymentState,
    PendingModel,
    PendingUser,
    PendingGroup,
    PendingMemberships,
)
from ldf_adapter.approval.db import generic


logger = logging.getLogger(__name__)


T = TypeVar("T")


def pytype_to_sqltype(pytype: Type) -> str:
    """Return the sql type for any given python type."""

    def get_type_name(pytype):
        try:
            return pytype.__repr__()
        except TypeError:
            return pytype.__name__

    type_name = get_type_name(pytype)
    module_name = pytype.__module__
    # types from typing module supported by sqlite
    if module_name == "typing":
        if type_name.startswith("typing.Optional") or type_name.startswith(
            "typing.Union"
        ):
            # Union only when it's used to represent an optional field
            # in python 3.7, __repr__() for the Optional type returns the Union representation
            # in python >= 3.9, __repr__() for a Union of a type and NoneType returns the Optional representation
            if (
                len(pytype.__args__) == 2
                and get_type_name(pytype.__args__[0]) == "NoneType"
            ):
                return pytype_to_sqltype(pytype.__args__[1])
            elif (
                len(pytype.__args__) == 2
                and get_type_name(pytype.__args__[1]) == "NoneType"
            ):
                return pytype_to_sqltype(pytype.__args__[0])
            else:
                return "text"
        elif type_name.startswith("typing.Literal"):
            return "text"
        elif type_name.startswith("typing.Dict"):
            return "dict"
        elif type_name.startswith("typing.List"):
            return "list"
        elif type_name.startswith("typing.Set"):
            return "list"
        elif type_name.startswith("typing.ByteString"):
            return "blob"
        else:
            return "text"  # default type
    if module_name == "builtins":
        # builtin types supported by sqlite
        if type_name == "str":
            return "text"
        if type_name == "int":
            return "integer"
        if type_name == "float":
            return "real"
        if type_name == "bytes":
            return "blob"
        if type_name == "NoneType":
            return "null"
        # additionally defined types with custom adapters and converters
        if type_name == "dict":
            return "dict"
        if type_name == "list":
            return "list"
        if type_name == "bool":
            return "boolean"
    # more additionally defined types with custom adapters and converters from other modules
    if type_name == "DeploymentState":
        return "DeploymentState"
    # default type
    return "text"


def sql_command_create_table(
    data_model: Type[PendingModel], table_name: str, primary_key: Union[str, List[str]]
) -> str:
    """Return an sql command for creating a table for a given data model.

    Args:
        data_model (Type[PendingModel]): dataclass containing the fields that will become the table columns
        table_name (str): the table name
        primary_key (Union[str, List[str]]): name(s) of column(s) to be used as primary key

    Returns:
        str: a string representation of the sql command
    """
    columns_with_types = ", ".join(
        [
            " ".join([field.name, pytype_to_sqltype(field.type)])
            for field in fields(data_model)
        ]
    )
    if isinstance(primary_key, List):
        primary_key = ", ".join(primary_key)
    return (
        f"create table if not exists {table_name}"
        f"({columns_with_types}, primary key ({primary_key}))"
    )


def sql_command_insert(data_model: Type[PendingModel], table_name: str) -> str:
    """Return an sql command for inserting an entry of a given data model into a table.

    The sql command will require the following arguments:
    - the values of the dataclass fields to be inserted

    Args:
        data_model (Type[PendingModel]): dataclass containing the fields that are the same as the table columns
        table_name (str): the table name

    Returns:
        str: a string representation of the sql command
    """
    column_names = ", ".join([field.name for field in fields(data_model)])
    column_values = ",".join(["?" for _ in fields(data_model)])
    return f"insert into {table_name}({column_names}) values ({column_values})"


def sql_command_update(
    table_name: str, columns: List[str], key: Union[str, List[str]]
) -> str:
    """Return an sql command for updating a set of fields for an entry in a given table.

    The sql command will require the following arguments, in this order:
    - the new values for each column
    - the key value(s) of the entry to be updated

    Args:
        table_name (str): the table name
        columns (list[str]): the columns to update
        key (list | str): the key to search on for update; a column name or a list of column names

    Returns:
        str: a string representation of the sql command
    """
    column_names = ", ".join([f"{col} = ?" for col in columns])
    condition = (
        " and ".join([f"{k} = ?" for k in key])
        if isinstance(key, list)
        else f"{key} = ?"
    )
    return f"update {table_name} set {column_names} where {condition}"


def sql_command_remove(table_name: str, key: Union[str, List[str]]) -> str:
    """Return an sql command for removing an entry from a table.

    The sql command will require the following arguments:
    - the values for given key(s) of the entry to be removed

    Args:
        table_name (str): the table name
        key (list | str): the key to search on for removal; a column name or a list of column names

    Returns:
        str: a string representation of the sql command
    """
    condition = (
        " and ".join([f"{k} = ?" for k in key])
        if isinstance(key, list)
        else f"{key} = ?"
    )
    return f"delete from {table_name} where {condition}"


def sql_command_select(table_name: str, key: Union[str, List[str]]) -> str:
    """Return an sql command for selecting an entry from a table.

    The sql command will require the following arguments:
    - the values for given key(s) of the entry to be selected

    Args:
        table_name (str): the table name
        key (list | str): the key to search on for selection; a column name or a list of column names

    Returns:
        str: a string representation of the sql command
    """
    condition = (
        " and ".join([f"{k} = ?" for k in key])
        if isinstance(key, list)
        else f"{key} = ?"
    )
    return f"select * from {table_name} where {condition}"


# register the functions for manipulating custom types in sqlite db
sqlite3.register_adapter(dict, lambda x: json.dumps(x).encode("utf-8"))
sqlite3.register_converter("dict", lambda x: json.loads(x.decode("utf-8")))

sqlite3.register_adapter(list, lambda x: json.dumps(x).encode("utf-8"))
sqlite3.register_converter("list", lambda x: json.loads(x.decode("utf-8")))

sqlite3.register_adapter(DeploymentState, lambda x: x.name.encode("utf-8"))
sqlite3.register_converter(
    "DeploymentState", lambda x: DeploymentState.from_string(x.decode("utf-8"))
)

sqlite3.register_adapter(bool, int)
sqlite3.register_converter("boolean", lambda v: bool(int(v)))


class SQLiteConnector:
    def __init__(self, location: str) -> None:
        self.location = location
        Path(self.location).parent.mkdir(mode=0o700, parents=True, exist_ok=True)

    def exists(self) -> bool:
        return Path(self.location).exists()

    def connect(self) -> None:
        """Connect to the DB.
        Initialise self.connection object.
        """
        try:
            self.connection = sqlite3.connect(
                self.location, detect_types=sqlite3.PARSE_DECLTYPES
            )
        except sqlite3.Error as ex:
            message = (
                f"Could not connect to sqlite db at location {self.location}: {ex}"
            )
            logger.error(message)
            raise FatalError(message=message)

    def create(
        self,
        data_model: Type[PendingModel],
        table_name: str,
        primary_key: Union[str, List[str]],
    ) -> None:
        """Create a table for a given data model.

        Args:
            data_model (Type[PendingModel]): dataclass containing the fields that will become the table columns
            table_name (str): the table name
            primary_key (Union[str, List[str]]): name(s) of column(s) to be used as primary key
        """
        try:
            with self.connection:  # con.commit() is called automatically afterwards on success
                logger.debug("Creating table: %s", table_name)
                self.connection.execute(
                    sql_command_create_table(
                        data_model=data_model,
                        table_name=table_name,
                        primary_key=primary_key,
                    )
                )
                logger.debug("Successfully created table '%s'.", table_name)
        except sqlite3.Error as ex:
            message = f"Failed to create table {table_name}: {ex}"
            logger.error(message)
            raise FatalError(message=message)

    def insert(
        self, data_model: Type[PendingModel], table_name: str, entry: tuple
    ) -> bool:
        """Insert an entry of a given data model into a table.

        Args:
            data_model (Type[PendingModel]): dataclass containing the fields that are the same as the table columns
            table_name (str): the table name
            entry (tuple): the values of the dataclass fields to be inserted

        Returns:
            bool: False if an entry already existed for given key, True if a new entry was inserted.
        """
        try:
            with self.connection:
                logger.debug("Inserting to table: %s", table_name)
                self.connection.execute(
                    sql_command_insert(data_model=data_model, table_name=table_name),
                    entry,
                )
                logger.debug(
                    "Successfully inserted to table '%s': %s", table_name, entry
                )
            return True
        except sqlite3.IntegrityError as ex:
            if ex.args[0].startswith("UNIQUE constraint failed: "):
                logger.info(f"Entry {entry} already exists in table {table_name}.")
                return False
            raise ex
        except sqlite3.Error as ex:
            message = f"Failed to insert entry to table {table_name}: {ex}"
            logger.error(message)
            raise Failure(message=message)

    def update(
        self,
        table_name: str,
        columns: List[str],
        key: Union[str, List[str]],
        entry: tuple,
    ) -> None:
        """Update an entry in a table.

        Args:
            table_name (str): the table name
            columns (list[str]): the columns to update
            key (str | list[str]): the key to search on for update
            entry (tuple): the values of the columns to be updated, and the value(s) of the key(s)
                           to be used for search, in this order
        """
        try:
            with self.connection:
                logger.debug("Updating entry in table: %s", table_name)
                self.connection.execute(
                    sql_command_update(table_name=table_name, columns=columns, key=key),
                    entry,
                )
                logger.debug(
                    "Successfully updated entry in table '%s': %s", table_name, entry
                )
        except sqlite3.Error as ex:
            message = f"Failed to update table {table_name}: {ex}"
            logger.error(message)
            raise Failure(message=message)

    def select(
        self,
        data_model: Type[T],
        table_name: str,
        key: Union[str, List[str]],
        value: tuple,
    ) -> Optional[T]:
        """Get an entry from a table.

        Args:
            data_model (Type[T]): data type of the entry to be returned
            table_name (str): the table name
            key (Union[str, List[str]]): name(s) of column(s) to be used as search key
            value (tuple): the value(s) of key(s) for which to search

        Returns:
            Optional[T]: the (first) entry found or None if no entry was found
        """
        try:
            with self.connection:
                logger.debug("Getting from table: %s", table_name)
                result = self.connection.execute(
                    sql_command_select(table_name=table_name, key=key), value
                ).fetchall()
                logger.debug(
                    "Successfully got entry from table '%s' for value %s.",
                    table_name,
                    value,
                )
                if len(result) == 0:
                    return None
                if len(result) > 1:
                    logger.warning(
                        "Multiple entries found in table %s for value: %s",
                        table_name,
                        value,
                    )
                return data_model(*result[0])
        except sqlite3.Error as ex:
            message = (
                f"Failed to get entry from table {table_name} for value {value}: {ex}"
            )
            logger.error(message)
            raise Failure(message=message)

    def delete(self, table_name: str, key: Union[str, List[str]], value: tuple) -> None:
        """Remove an entry from a table.

        Args:
            table_name (str): the table name
            key (Union[str, List[str]]): name(s) of column(s) to be used search key
            value (tuple): value(s) of key(s) for entry to be removed
        """
        try:
            with self.connection:
                logger.debug("Removing from table: %s", table_name)
                self.connection.execute(
                    sql_command_remove(table_name=table_name, key=key), value
                )
                logger.debug(
                    "Successfully removed entry %s from table '%s'.", value, table_name
                )
        except sqlite3.Error as ex:
            message = f"Failed to remove entry {value} from table {table_name}: {ex}"
            logger.error(message)
            raise Failure(message=message)


class PendingDB(generic.PendingDB):
    """Implementation of PendingDB with sqlite3."""

    def __init__(self) -> None:
        """Initialise sqlite-based DB for managing users and groups pending approval."""
        self.connector = SQLiteConnector(CONFIG.approval.user_db_location)
        if not self.connector.exists():
            self.connector.connect()
            logger.debug(
                "No sqlite DB found at %s, creating it...",
                CONFIG.approval.user_db_location,
            )
            self.connector.create(
                data_model=PendingUser,
                table_name="pending_users",
                primary_key="unique_id",
            )
            self.connector.create(
                data_model=PendingGroup, table_name="pending_groups", primary_key="name"
            )
            self.connector.create(
                data_model=PendingMemberships,
                table_name="pending_memberships",
                primary_key="unique_id",
            )
        else:
            logger.debug(
                "Existing sqlite DB found at %s. Loading pending data from it.",
                CONFIG.approval.user_db_location,
            )
            self.connector.connect()

    def add_user(self, user: PendingUser) -> bool:
        """Add a new entry for a user. Returns False if entry already exists."""
        return self.connector.insert(
            data_model=PendingUser,
            table_name="pending_users",
            entry=tuple(getattr(user, field.name) for field in fields(PendingUser)),
        )

    def remove_user(self, unique_id: str) -> None:
        """Remove user entry for unique_id."""
        self.connector.delete(
            table_name="pending_users", key="unique_id", value=(unique_id,)
        )

    def get_user(self, unique_id: str) -> Optional[PendingUser]:
        """Get a user entry that is up for approval by the user's unique_id."""
        return self.connector.select(
            data_model=PendingUser,
            table_name="pending_users",
            key="unique_id",
            value=(unique_id,),
        )

    def get_user_by_username(self, username: str) -> Optional[PendingUser]:
        """Get an entry by the local username."""
        return self.connector.select(
            data_model=PendingUser,
            table_name="pending_users",
            key="username",
            value=(username,),
        )

    def reject_user(self, unique_id: str) -> None:
        """Change the state of user given by unique_id to REJECTED."""
        self.connector.update(
            table_name="pending_users",
            columns=["state"],
            key="unique_id",
            entry=(DeploymentState.REJECTED, unique_id),
        )

    def notify_user(self, unique_id: str) -> None:
        """Change the state of user given by unique_id to NOTIFIED."""
        self.connector.update(
            table_name="pending_users",
            columns=["state"],
            key="unique_id",
            entry=(DeploymentState.NOTIFIED, unique_id),
        )

    def add_group(self, group: PendingGroup) -> bool:
        """Add a new entry for a group. Returns False if entry already exists."""
        return self.connector.insert(
            data_model=PendingGroup,
            table_name="pending_groups",
            entry=tuple(getattr(group, field.name) for field in fields(PendingGroup)),
        )

    def remove_group(self, name: str) -> None:
        """Remove group entry for name."""
        self.connector.delete(table_name="pending_groups", key="name", value=(name,))

    def get_group(self, name: str) -> Optional[PendingGroup]:
        """Get an entry that is up for approval by the group's name."""
        return self.connector.select(
            data_model=PendingGroup,
            table_name="pending_groups",
            key="name",
            value=(name,),
        )

    def notify_group(self, name: str) -> None:
        """Change the state of group given by name to NOTIFIED."""
        self.connector.update(
            table_name="pending_groups",
            columns=["state"],
            key="name",
            entry=(DeploymentState.NOTIFIED, name),
        )

    def add_memberships(self, membership: PendingMemberships) -> bool:
        """Add a new entry for a user's pending group memberships. Returns False if entry already exists."""
        return self.connector.insert(
            data_model=PendingMemberships,
            table_name="pending_memberships",
            entry=tuple(
                getattr(membership, field.name) for field in fields(PendingMemberships)
            ),
        )

    def update_memberships(self, membership: PendingMemberships) -> None:
        """Update a user's pending group memberships by replacing them with the given memberships."""
        columns = [
            field.name
            for field in fields(PendingMemberships)
            if field.name != "unique_id"
        ]
        self.connector.update(
            table_name="pending_memberships",
            columns=columns,
            key="unique_id",
            entry=tuple(
                [getattr(membership, col) for col in columns] + [membership.unique_id]
            ),
        )

    def remove_memberships(self, unique_id: str) -> None:
        """Remove all pending group memberships of a given user."""
        self.connector.delete(
            table_name="pending_memberships", key="unique_id", value=(unique_id,)
        )

    def get_memberships(self, unique_id: str) -> Optional[PendingMemberships]:
        """Get a given user's pending group memberships."""
        return self.connector.select(
            data_model=PendingMemberships,
            table_name="pending_memberships",
            key="unique_id",
            value=(unique_id,),
        )

    def notify_memberships(self, unique_id: str) -> None:
        """Change the state of user's memberships given by unique_id to NOTIFIED."""
        self.connector.update(
            table_name="pending_memberships",
            columns=["state"],
            key="unique_id",
            entry=(DeploymentState.NOTIFIED, unique_id),
        )
